#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

from swimai.structures import Text
from swimai.structures._structs import _Absent
from swimai.warp._warp import _Envelope, _SyncRequestForm, _SyncedResponseForm, _LinkedResponseForm, _EventMessageForm, \
    _CommandMessageForm, _SyncRequest, _SyncedResponse, _LinkedResponse, _CommandMessage, _EventMessage, _LinkRequestForm, \
    _UnlinkedResponseForm, _LinkRequest, _UnlinkedResponse


class TestEnvelopes(unittest.TestCase):

    def test_resolve_form_sync(self):
        # Given
        tag = 'sync'
        # When
        actual = _Envelope._resolve_form(tag)
        # Then
        self.assertIsInstance(actual, _SyncRequestForm)

    def test_resolve_form_synced(self):
        # Given
        tag = 'synced'
        # When
        actual = _Envelope._resolve_form(tag)
        # Then
        self.assertIsInstance(actual, _SyncedResponseForm)

    def test_resolve_form_link(self):
        # Given
        tag = 'link'
        # When
        actual = _Envelope._resolve_form(tag)
        # Then
        self.assertIsInstance(actual, _LinkRequestForm)

    def test_resolve_form_linked(self):
        # Given
        tag = 'linked'
        # When
        actual = _Envelope._resolve_form(tag)
        # Then
        self.assertIsInstance(actual, _LinkedResponseForm)

    def test_resolve_form_unlinked(self):
        # Given
        tag = 'unlinked'
        # When
        actual = _Envelope._resolve_form(tag)
        # Then
        self.assertIsInstance(actual, _UnlinkedResponseForm)

    def test_resolve_form_event(self):
        # Given
        tag = 'event'
        # When
        actual = _Envelope._resolve_form(tag)
        # Then
        self.assertIsInstance(actual, _EventMessageForm)

    def test_resolve_form_command(self):
        # Given
        tag = 'command'
        # When
        actual = _Envelope._resolve_form(tag)
        # Then
        self.assertIsInstance(actual, _CommandMessageForm)

    def test_resolve_form_invalid(self):
        # Given
        tag = 'this_is_not_a_valid_form'
        # When
        with self.assertRaises(TypeError) as error:
            _Envelope._resolve_form(tag)
        # Then
        message = error.exception.args[0]
        self.assertEqual(message, 'Invalid form tag: this_is_not_a_valid_form')

    def test_sync_request_empty_body(self):
        # Given
        node_uri = 'foo__sync_node'
        lane_uri = 'bar__sync_lane'
        # When
        actual = _SyncRequest(node_uri, lane_uri)
        # Then
        self.assertEqual('foo__sync_node', actual._node_uri)
        self.assertEqual('bar__sync_lane', actual._lane_uri)
        self.assertEqual(0.0, actual._prio)
        self.assertEqual(0.0, actual._rate)
        self.assertEqual('sync', actual._tag)
        self.assertEqual('foo__sync_node/bar__sync_lane', actual._route)
        self.assertEqual(_Absent._get_absent(), actual._body)
        self.assertIsInstance(actual._form, _SyncRequestForm)

    def test_sync_request_existing_body(self):
        # Given
        node_uri = 'foo_sync_node'
        lane_uri = 'bar_sync_lane'
        body = Text.create_from('Sync_Body')
        priority = 3.5
        rate = 2.0
        # When
        actual = _SyncRequest(node_uri, lane_uri, priority, rate, body=body)
        # Then
        self.assertEqual('foo_sync_node', actual._node_uri)
        self.assertEqual('bar_sync_lane', actual._lane_uri)
        self.assertEqual(3.5, actual._prio)
        self.assertEqual(2.0, actual._rate)
        self.assertEqual('sync', actual._tag)
        self.assertEqual('foo_sync_node/bar_sync_lane', actual._route)
        self.assertEqual(body, actual._body)
        self.assertIsInstance(actual._form, _SyncRequestForm)

    def test_synced_response_empty_body(self):
        # Given
        node_uri = 'foo_synced_node'
        lane_uri = 'bar_synced_lane'
        # When
        actual = _SyncedResponse(node_uri, lane_uri)
        # Then
        self.assertEqual('foo_synced_node', actual._node_uri)
        self.assertEqual('bar_synced_lane', actual._lane_uri)
        self.assertEqual('synced', actual._tag)
        self.assertEqual('foo_synced_node/bar_synced_lane', actual._route)
        self.assertEqual(_Absent._get_absent(), actual._body)
        self.assertIsInstance(actual._form, _SyncedResponseForm)

    def test_synced_response_existing_body(self):
        # Given
        node_uri = 'foo_synced_node'
        lane_uri = 'bar_synced_lane'
        body = Text.create_from('Synced_Body')
        # When
        actual = _SyncedResponse(node_uri, lane_uri, body=body)
        # Then
        self.assertEqual('foo_synced_node', actual._node_uri)
        self.assertEqual('bar_synced_lane', actual._lane_uri)
        self.assertEqual('synced', actual._tag)
        self.assertEqual('foo_synced_node/bar_synced_lane', actual._route)
        self.assertEqual(body, actual._body)
        self.assertIsInstance(actual._form, _SyncedResponseForm)

    def test_link_request_empty_body(self):
        # Given
        node_uri = 'foo_link_node'
        lane_uri = 'bar_link_lane'
        # When
        actual = _LinkRequest(node_uri, lane_uri)
        # Then
        self.assertEqual('foo_link_node', actual._node_uri)
        self.assertEqual('bar_link_lane', actual._lane_uri)
        self.assertEqual('link', actual._tag)
        self.assertEqual('foo_link_node/bar_link_lane', actual._route)
        self.assertEqual(_Absent._get_absent(), actual._body)
        self.assertIsInstance(actual._form, _LinkRequestForm)

    def test_link_request_existing_body(self):
        # Given
        node_uri = 'foo_link_node'
        lane_uri = 'bar_link_lane'
        body = Text.create_from('Link_Body')
        priority = 13.1
        rate = 11.3
        # When
        actual = _LinkRequest(node_uri, lane_uri, priority, rate, body)
        # Then
        self.assertEqual('foo_link_node', actual._node_uri)
        self.assertEqual('bar_link_lane', actual._lane_uri)
        self.assertEqual('link', actual._tag)
        self.assertEqual('foo_link_node/bar_link_lane', actual._route)
        self.assertEqual(body, actual._body)
        self.assertIsInstance(actual._form, _LinkRequestForm)

    def test_linked_response_empty_body(self):
        # Given
        node_uri = 'foo_linked_node'
        lane_uri = 'bar_linked_lane'
        # When
        actual = _LinkedResponse(node_uri, lane_uri)
        # Then
        self.assertEqual('foo_linked_node', actual._node_uri)
        self.assertEqual('bar_linked_lane', actual._lane_uri)
        self.assertEqual('linked', actual._tag)
        self.assertEqual('foo_linked_node/bar_linked_lane', actual._route)
        self.assertEqual(_Absent._get_absent(), actual._body)
        self.assertIsInstance(actual._form, _LinkedResponseForm)

    def test_linked_response_existing_body(self):
        # Given
        node_uri = 'foo_linked_node'
        lane_uri = 'bar_linked_lane'
        body = Text.create_from('Linked_Body')
        priority = 1.13
        rate = 3.11
        # When
        actual = _LinkedResponse(node_uri, lane_uri, priority, rate, body)
        # Then
        self.assertEqual('foo_linked_node', actual._node_uri)
        self.assertEqual('bar_linked_lane', actual._lane_uri)
        self.assertEqual('linked', actual._tag)
        self.assertEqual('foo_linked_node/bar_linked_lane', actual._route)
        self.assertEqual(body, actual._body)
        self.assertIsInstance(actual._form, _LinkedResponseForm)

    def test_unlinked_response_empty_body(self):
        # Given
        node_uri = 'foo_unlinked_node'
        lane_uri = 'bar_unlinked_lane'
        # When
        actual = _UnlinkedResponse(node_uri, lane_uri)
        # Then
        self.assertEqual('foo_unlinked_node', actual._node_uri)
        self.assertEqual('bar_unlinked_lane', actual._lane_uri)
        self.assertEqual('unlinked', actual._tag)
        self.assertEqual('foo_unlinked_node/bar_unlinked_lane', actual._route)
        self.assertEqual(_Absent._get_absent(), actual._body)
        self.assertIsInstance(actual._form, _UnlinkedResponseForm)

    def test_unlinked_response_existing_body(self):
        # Given
        node_uri = 'foo_unlinked_node'
        lane_uri = 'bar_unlinked_lane'
        body = Text.create_from('Unlinked_Body')
        priority = 5.11
        rate = 6.13
        # When
        actual = _UnlinkedResponse(node_uri, lane_uri, priority, rate, body)
        # Then
        self.assertEqual('foo_unlinked_node', actual._node_uri)
        self.assertEqual('bar_unlinked_lane', actual._lane_uri)
        self.assertEqual('unlinked', actual._tag)
        self.assertEqual('foo_unlinked_node/bar_unlinked_lane', actual._route)
        self.assertEqual(body, actual._body)
        self.assertIsInstance(actual._form, _UnlinkedResponseForm)

    def test_command_message_empty_body(self):
        # Given
        node_uri = 'foo_command_node'
        lane_uri = 'bar_command_lane'
        # When
        actual = _CommandMessage(node_uri, lane_uri)
        # Then
        self.assertEqual('foo_command_node', actual._node_uri)
        self.assertEqual('bar_command_lane', actual._lane_uri)
        self.assertEqual('command', actual._tag)
        self.assertEqual('foo_command_node/bar_command_lane', actual._route)
        self.assertEqual(_Absent._get_absent(), actual._body)
        self.assertIsInstance(actual._form, _CommandMessageForm)

    def test_command_message_existing_body(self):
        # Given
        node_uri = 'foo_command_node'
        lane_uri = 'bar_command_lane'
        body = Text.create_from('Command_Body')
        # When
        actual = _CommandMessage(node_uri, lane_uri, body)
        # Then
        self.assertEqual('foo_command_node', actual._node_uri)
        self.assertEqual('bar_command_lane', actual._lane_uri)
        self.assertEqual('command', actual._tag)
        self.assertEqual('foo_command_node/bar_command_lane', actual._route)
        self.assertEqual(body, actual._body)
        self.assertIsInstance(actual._form, _CommandMessageForm)

    def test_event_message_empty_body(self):
        # Given
        node_uri = 'foo_event_node'
        lane_uri = 'bar_event_lane'
        # When
        actual = _EventMessage(node_uri, lane_uri)
        # Then
        self.assertEqual('foo_event_node', actual._node_uri)
        self.assertEqual('bar_event_lane', actual._lane_uri)
        self.assertEqual('event', actual._tag)
        self.assertEqual('foo_event_node/bar_event_lane', actual._route)
        self.assertEqual(_Absent._get_absent(), actual._body)
        self.assertIsInstance(actual._form, _EventMessageForm)

    def test_event_message_existing_body(self):
        # Given
        node_uri = 'foo_event_node'
        lane_uri = 'bar_event_lane'
        body = Text.create_from('Event_Body')
        # When
        actual = _EventMessage(node_uri, lane_uri, body)
        # Then
        self.assertEqual('foo_event_node', actual._node_uri)
        self.assertEqual('bar_event_lane', actual._lane_uri)
        self.assertEqual('foo_event_node/bar_event_lane', actual._route)
        self.assertEqual('event', actual._tag)
        self.assertEqual(body, actual._body)
        self.assertIsInstance(actual._form, _EventMessageForm)



#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import websockets

from enum import Enum
from swimai.warp._warp import _Envelope
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ._downlinks._downlinks import _DownlinkModel
    from ._downlinks._downlinks import _DownlinkView


class _ConnectionPool:

    def __init__(self) -> None:
        self.__connections = dict()

    @property
    def _size(self) -> int:
        return len(self.__connections)

    async def _get_connection(self, host_uri: str) -> '_WSConnection':
        """
        Return a WebSocket connection to the given Host URI. If it is a new
        host or the existing connection is closing, create a new connection.

        :param host_uri:        - URI of the connection host.
        :return:                - WebSocket connection.
        """
        connection = self.__connections.get(host_uri)

        if connection is None or connection.status == _ConnectionStatus.CLOSED:
            connection = _WSConnection(host_uri)
            self.__connections[host_uri] = connection

        return connection

    async def _remove_connection(self, host_uri: str) -> None:
        """
        Remove a connection from the pool.

        :param host_uri:        - URI of the connection host.
        """
        connection = self.__connections.get(host_uri)

        if connection:
            self.__connections.pop(host_uri)
            await connection._close()

    async def _add_downlink_view(self, downlink_view: '_DownlinkView') -> None:
        """
        Subscribe a downlink view to a connection from the pool.

        :param downlink_view:   - Downlink view to subscribe to a connection.
        """
        host_uri = downlink_view._host_uri
        connection = await self._get_connection(host_uri)
        downlink_view._connection = connection

        await connection._subscribe(downlink_view)

    async def _remove_downlink_view(self, downlink_view: '_DownlinkView') -> None:
        """
        Unsubscribe a downlink view from a connection from the pool.

        :param downlink_view:   - Downlink view to unsubscribe from a connection.
        """
        connection: '_WSConnection'

        host_uri = downlink_view._host_uri
        connection = self.__connections.get(host_uri)

        if connection:
            await connection._unsubscribe(downlink_view)

            if connection.status == _ConnectionStatus.CLOSED:
                await self._remove_connection(host_uri)


class _WSConnection:

    def __init__(self, host_uri: str) -> None:
        self.host_uri = host_uri
        self.connected = asyncio.Event()
        self.websocket = None
        self.status = _ConnectionStatus.CLOSED

        self.__subscribers = _DownlinkManagerPool()

    async def _open(self) -> None:
        if self.status == _ConnectionStatus.CLOSED:
            self.status = _ConnectionStatus.CONNECTING

            try:
                self.websocket = await websockets.connect(self.host_uri)
            except Exception as error:
                self.status = _ConnectionStatus.CLOSED
                raise error

            self.status = _ConnectionStatus.IDLE
            self.connected.set()

    async def _close(self) -> None:
        if self.status != _ConnectionStatus.CLOSED:
            self.status = _ConnectionStatus.CLOSED

            if self.websocket:
                self.websocket.close_timeout = 0.1
                await self.websocket.close()
                self.connected.clear()

    def _has_subscribers(self) -> bool:
        """
        Check if the connection has any subscribers.

        :return:        - True if there are subscribers. False otherwise.
        """
        return self.__subscribers._size > 0

    async def _subscribe(self, downlink_view: '_DownlinkView') -> None:
        """
        Add a downlink view to the subscriber list of the current connection.
        If this is the first subscriber, open the connection.

        :param downlink_view:   - Downlink view to add to the subscribers.
        """
        if self.__subscribers._size == 0:
            await self._open()

        await self.__subscribers._register_downlink_view(downlink_view)

    async def _unsubscribe(self, downlink_view: '_DownlinkView') -> None:
        """
        Remove a downlink view from the subscriber list of the current connection.
        If there are no other subscribers, close the connection.

        :param downlink_view:   - Downlink view to remove from the subscribers.
        """

        await self.__subscribers._deregister_downlink_view(downlink_view)
        if not self._has_subscribers():
            await self._close()

    async def _send_message(self, message: str) -> None:
        """
        Send a string message to the host using a WebSocket connection.
        If the WebSocket connection to the host is not open, open it.

        :param message:         - String message to send to the remote agent.
        """
        if self.websocket is None or self.status == _ConnectionStatus.CLOSED:
            await self._open()

        await self.connected.wait()
        await self.websocket.send(message)

    async def _wait_for_messages(self) -> None:
        """
        Wait for messages from the remote agent and propagate them
        to all subscribers.
        """

        if self.status == _ConnectionStatus.IDLE:
            self.status = _ConnectionStatus.RUNNING
            try:
                while self.status == _ConnectionStatus.RUNNING:
                    message = await self.websocket.recv()
                    response = await _Envelope._parse_recon(message)
                    await self.__subscribers._receive_message(response)
            finally:
                await self._close()


class _ConnectionStatus(Enum):
    CLOSED = 0
    CONNECTING = 1
    IDLE = 2
    RUNNING = 3


class _DownlinkManagerPool:

    def __init__(self) -> None:
        self.__downlink_managers = dict()

    @property
    def _size(self) -> int:
        return len(self.__downlink_managers)

    async def _register_downlink_view(self, downlink_view: '_DownlinkView') -> None:
        """
        Add a downlink view to a downlink manager from the pool with the given node and lane URIs.
        If a downlink manager is not yet created for the given node and lane, create it and add the downlink view.


        :param downlink_view:   - Downlink view to add to a corresponding downlink manager.
        """
        downlink_manager = self.__downlink_managers.get(downlink_view.route)

        if downlink_manager is None:
            downlink_manager = _DownlinkManager(downlink_view._connection)
            self.__downlink_managers[downlink_view.route] = downlink_manager

        await downlink_manager._add_view(downlink_view)

    async def _deregister_downlink_view(self, downlink_view: '_DownlinkView') -> None:
        """
        Remove a downlink view from the corresponding downlink manager if it exists.
        If it is the last downlink view in the given manager, remove the manager from the pool.

        :param downlink_view:   - Downlink view to remove from the corresponding downlink manager.
        """
        downlink_manager: _DownlinkManager

        if downlink_view.route in self.__downlink_managers:
            downlink_manager = self.__downlink_managers.get(downlink_view.route)
            await downlink_manager._remove_view(downlink_view)

            if downlink_manager._view_count == 0:
                self.__downlink_managers.pop(downlink_view.route)

    async def _receive_message(self, message: '_Envelope') -> None:
        """
        Route a received message for the given host URI to the downlink manager for the corresponding
        node and lane URIs.

        :param message:         - Message received from the remote agent.
        """
        downlink_manager: _DownlinkManager

        downlink_manager = self.__downlink_managers.get(message._route)
        if downlink_manager:
            await downlink_manager._receive_message(message)


class _DownlinkManager:

    def __init__(self, connection: '_WSConnection') -> None:
        self.connection = connection
        self.status = _DownlinkManagerStatus.CLOSED
        self.downlink_model = None
        self.registered_classes = dict()
        self.strict = False

        self.__downlink_views = dict()

    @property
    def _view_count(self) -> int:
        return len(self.__downlink_views)

    @property
    def _is_open(self) -> bool:
        return self.status == _DownlinkManagerStatus.OPEN

    async def _open(self) -> None:
        self.downlink_model: _DownlinkModel

        if self.status == _DownlinkManagerStatus.CLOSED:
            self.status = _DownlinkManagerStatus.OPENING
            self.downlink_model._open()
            await self.downlink_model._establish_downlink()
            self.status = _DownlinkManagerStatus.OPEN

    async def _close(self) -> None:
        self.downlink_model: _DownlinkModel

        if self.status != _DownlinkManagerStatus.CLOSED:
            self.status = _DownlinkManagerStatus.CLOSED
            self.downlink_model._close()

    async def _init_downlink_model(self, downlink_view: '_DownlinkView') -> None:
        """
        Initialise a downlink model to the specified node and lane of the remote agent.

        :param downlink_view:       - Downlink view with the information about the remote agent.
        """
        self.downlink_model = await downlink_view._create_downlink_model(self)
        self.downlink_model.connection = self.connection

    async def _add_view(self, downlink_view: '_DownlinkView') -> None:
        """
        Add a downlink view to the manager. If a downlink model is not yet created, create it and open it.

        :param downlink_view:       - Downlink view to add to the manager.
        """
        if self.downlink_model is None:
            await self._init_downlink_model(downlink_view)

        await downlink_view._register_manager(self)

        if self._view_count == 0:
            await self._open()

        self.__downlink_views[hash(downlink_view)] = downlink_view

    async def _remove_view(self, downlink_view: '_DownlinkView') -> None:
        """
        Remove a downlink view from the manager. If it is the last view associated with the manager,
        close the manager.

        :param downlink_view:       - Downlink view to remove from the manager.
        """
        if hash(downlink_view) in self.__downlink_views:
            self.__downlink_views.pop(hash(downlink_view))

            if self._view_count == 0:
                await self._close()

    async def _receive_message(self, message: '_Envelope') -> None:
        """
        Send a received message to the downlink model.

        :param message:             - Received message from the remote agent.
        """
        self.downlink_model: _DownlinkModel

        await self.downlink_model._receive_message(message)

    async def _subscribers_did_set(self, current_value: Any, old_value: Any) -> None:
        """
        Execute the `did_set` method of all value downlink views of the downlink manager.

        :param current_value:       - The new value of the downlink.
        :param old_value:           - The previous value of the downlink.
        """
        for view in self.__downlink_views.values():
            await view._execute_did_set(current_value, old_value)

    async def _subscribers_on_event(self, event: Any) -> None:
        """
        Execute the `on_event` method of all event downlink views of the downlink manager.

        :param event:       - Event from the remote lane.
        """

        for view in self.__downlink_views.values():
            await view._execute_on_event(event)

    async def _subscribers_did_update(self, key: Any, new_value: Any, old_value: Any) -> None:
        """
        Execute the `did_update` method of all map downlink views of the downlink manager.

        :param key:                 - The key of the entry.
        :param new_value:           - The new value of entry.
        :param old_value:           - The previous value of the entry.
        """
        for view in self.__downlink_views.values():
            await view._execute_did_update(key, new_value, old_value)

    async def _subscribers_did_remove(self, key: Any, old_value: Any) -> None:
        """
         Execute the `did_remove` method of all map downlink views of the downlink manager.

         :param key:                 - The key of the entry.
         :param old_value:           - The previous value of the entry.
         """
        for view in self.__downlink_views.values():
            await view._execute_did_remove(key, old_value)

    def _close_views(self) -> None:
        """
        Set the status of all downlink views of the current manager to closed.
        """
        for view in self.__downlink_views.values():
            view._is_open = False


class _DownlinkManagerStatus(Enum):
    CLOSED = 0
    OPENING = 1
    OPEN = 2



#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
from asyncio import Future

from collections.abc import Callable
from abc import abstractmethod, ABC
from typing import TYPE_CHECKING, Any
from swimai.recon import Recon
from swimai.structures import Value, RecordConverter
from swimai.warp._warp import _SyncRequest, _CommandMessage, _Envelope, _LinkRequest
from .._utils import _URI
from ._utils import before_open, UpdateRequest, RemoveRequest, after_open, validate_callback

# Imports for type annotations
if TYPE_CHECKING:
    from .._swim_client import SwimClient
    from .._connections import _DownlinkManager


class _DownlinkModel(ABC):

    def __init__(self, client: 'SwimClient') -> None:
        self.client = client
        self.host_uri = None
        self.node_uri = None
        self.lane_uri = None
        self.task = None
        self.connection = None
        self.linked = asyncio.Event()
        self.downlink_manager = None

    @abstractmethod
    async def _establish_downlink(self) -> None:
        """
        Send a request to the remote agent to establish connection.
        """
        raise NotImplementedError

    @abstractmethod
    async def _receive_event(self, message: '_Envelope') -> None:
        """
        Handle an `event` response message from the remote agent.

        :param message:          - Event response message.
        """
        raise NotImplementedError

    @abstractmethod
    async def _receive_synced(self) -> None:
        """
        Handle a `synced` response message from the remote agent.
        """
        raise NotImplementedError

    def _open(self) -> '_DownlinkModel':
        self.task = self.client._schedule_task(self.connection._wait_for_messages)
        self.task.add_done_callback(self.__close_views)
        return self

    def _close(self) -> '_DownlinkModel':
        self.client._schedule_task(self.__close)
        return self

    async def _receive_message(self, message: '_Envelope') -> None:
        """
        Handle a response message from the remote agent.

        :param message:         - Response message.
        """
        if message._tag == 'linked':
            await self._receive_linked()
        elif message._tag == 'synced':
            await self._receive_synced()
        elif message._tag == 'event':
            await self._receive_event(message)
        elif message._tag == 'unlinked':
            await self._receive_unlinked(message)

    async def _receive_linked(self) -> None:
        """
        Handle a `linked` response message from the remote agent.
        """
        self.linked.set()

    async def _receive_unlinked(self, message: '_Envelope') -> None:
        """
        Handle an `unlinked` response message from the remote agent.

        :param message          - Response message.
        """
        if message._body._tag == 'laneNotFound':
            raise Exception(f'Lane "{self.lane_uri}" was not found on the remote agent!')

    async def __close(self) -> None:
        self.task.cancel()

    # noinspection PyUnusedLocal
    def __close_views(self, future: Future) -> None:
        if self.downlink_manager is not None:
            self.downlink_manager._close_views()


class _DownlinkView(ABC):

    def __init__(self, client: 'SwimClient') -> None:
        self._client = client
        self._host_uri = None
        self._node_uri = None
        self._lane_uri = None
        self._connection = None
        self._model = None
        self._downlink_manager = None
        self._is_open = False

        self.__registered_classes = dict()
        self.__deregistered_classes = set()
        self.__clear_classes = False
        self.__strict = False

    @property
    def route(self) -> str:
        return f'{self._node_uri}/{self._lane_uri}'

    @property
    def strict(self) -> bool:
        """
        The strict status is used when custom objects are received by the downlink. If the downlink is strict,
        a class of the same name, as the object received, must have been registered with the downlink. If the downlink
        is not strict, a custom object will be created dynamically.

        :return:                    - Whether or not the downlink is strict.
        """
        if self._downlink_manager is None:
            return self.__strict
        else:
            return self._downlink_manager.strict

    @strict.setter
    def strict(self, strict: bool) -> None:
        if self._downlink_manager is None:
            self.__strict = strict
        else:
            self._downlink_manager.strict = strict

    @property
    def registered_classes(self) -> dict:
        if self._downlink_manager is None:
            return self.__registered_classes
        else:
            return self._downlink_manager.registered_classes

    def open(self) -> '_DownlinkView':
        if not self._is_open:
            task = self._client._schedule_task(self._client._add_downlink_view, self)
            if task is not None:
                self._is_open = True

        return self

    def close(self) -> '_DownlinkView':
        if self._is_open:
            self._client._schedule_task(self._client._remove_downlink_view, self)
            self._is_open = False

        return self

    @before_open
    def set_host_uri(self, host_uri: str) -> '_DownlinkView':
        self._host_uri = _URI._normalise_warp_scheme(host_uri)
        return self

    @before_open
    def set_node_uri(self, node_uri: str) -> '_DownlinkView':
        self._node_uri = node_uri
        return self

    @before_open
    def set_lane_uri(self, lane_uri: str) -> '_DownlinkView':
        self._lane_uri = lane_uri
        return self

    def register_class(self, custom_class: Any) -> None:
        """
        Register a class with the downlink view. The registered classes are used for constructing objects when
        events with custom objects are received by the downlink.

        :param custom_class:        - Class to register.
        """
        self.__register_class(custom_class)

    def register_classes(self, classes_list: list) -> None:
        for custom_class in classes_list:
            self.__register_class(custom_class)

    def deregister_class(self, custom_class: Any) -> None:
        """
        Deregister a class from the downlink view and the downlink manager associated with it.

        :param custom_class:        - Class to deregister.
        """
        if self._downlink_manager is None:
            self.__registered_classes.pop(custom_class.__name__, None)
            self.__deregistered_classes.add(custom_class.__name__)
        else:
            self._downlink_manager.registered_classes.pop(custom_class.__name__, None)

    def deregister_classes(self, classes_list: list) -> None:
        for custom_class in classes_list:
            self.deregister_class(custom_class)

    def deregister_all_classes(self) -> None:
        """
        Deregister all classes from the downlink view and the downlink manager associated with it.
        """
        if self._downlink_manager is not None:
            self.__deregistered_classes.update(set(self._downlink_manager.registered_classes.keys()))
            self._downlink_manager.registered_classes.clear()
        else:
            self.__clear_classes = True
            self.__registered_classes.clear()

    @abstractmethod
    async def _register_manager(self, manager: '_DownlinkManager') -> None:
        """
        Register the current downlink view with a downlink manager.

        :param manager:             - Downlink manager for the current downlink view.
        """
        raise NotImplementedError

    @abstractmethod
    async def _create_downlink_model(self, downlink_manager: '_DownlinkManager') -> '_DownlinkModel':
        """
        Create a downlink model for the current downlink view.

        :param downlink_manager:    - Downlink manager of the current downlink view.
        :return:                    - Downlink model for the current downlink view.
        """
        raise NotImplementedError

    async def _initalise_model(self, manager: '_DownlinkManager', model: '_DownlinkModel') -> None:
        """
        Initialise the given downlink model and manager with the values of the current downlink view.

        :param manager:             - Downlink manager for the current downlink view.
        :param model:               - Downlink model for the current downlink view.
        """
        manager.registered_classes = self.registered_classes
        manager.strict = self.strict
        model.downlink_manager = manager
        model.host_uri = self._host_uri
        model.node_uri = self._node_uri
        model.lane_uri = self._lane_uri

    async def _assign_manager(self, manager: '_DownlinkManager') -> None:
        """
        Assign a downlink manager to the current downlink view and update the manager with the values
        of the downlink view.

        :param manager:             - Downlink manager to assign to the current downlink view.
        """
        self._model = manager.downlink_model
        manager.registered_classes.update(self.registered_classes)

        for klass in self.__deregistered_classes:
            manager.registered_classes.pop(klass)

        if self.__clear_classes:
            manager.registered_classes.clear()

        manager.strict = self.strict
        self._downlink_manager = manager

    def __register_class(self, custom_class: Any) -> None:
        try:
            custom_class()

            if self._downlink_manager is not None:
                self._downlink_manager.registered_classes[custom_class.__name__] = custom_class
            else:
                self.__registered_classes[custom_class.__name__] = custom_class
                self.__deregistered_classes.discard(custom_class.__name__)
        except Exception:
            raise Exception(
                f'Class "{custom_class.__name__}" must have a default constructor or default values for all arguments!')


class _EventDownlinkModel(_DownlinkModel):

    async def _establish_downlink(self) -> None:
        link_request = _LinkRequest(self.node_uri, self.lane_uri)
        await self.connection._send_message(await link_request._to_recon())

    async def _receive_event(self, message: _Envelope) -> None:
        converter = RecordConverter.get_converter()
        event = converter.record_to_object(message._body, self.downlink_manager.registered_classes,
                                           self.downlink_manager.strict)

        await self.downlink_manager._subscribers_on_event(event)

    async def _receive_synced(self) -> None:
        raise TypeError('Event downlink does not support synced responses!')


class _EventDownlinkView(_DownlinkView):

    def __init__(self, client: 'SwimClient') -> None:
        super().__init__(client)
        self._on_event_callback = None

    def on_event(self, function: Callable) -> '_EventDownlinkView':
        """
        Set the `on_event` callback of the current downlink view to a given function.

        :param function:   - Function to be called when an event is received by the downlink.
        :return:           - The current downlink view.
        """
        self._on_event_callback = validate_callback(function)
        return self

    async def _register_manager(self, manager: '_DownlinkManager') -> None:
        await self._assign_manager(manager)

    async def _create_downlink_model(self, downlink_manager: '_DownlinkManager') -> '_DownlinkModel':
        model = _EventDownlinkModel(self._client)
        await self._initalise_model(downlink_manager, model)
        return model

    # noinspection PyAsyncCall
    async def _execute_on_event(self, event: Any) -> None:
        """
        Execute the custom `on_event` callback of the current downlink view.

        :param event:       - The event received by the downlink.
        """
        if self._on_event_callback:
            self._client._schedule_task(self._on_event_callback, event)


class _ValueDownlinkModel(_DownlinkModel):

    def __init__(self, client: 'SwimClient') -> None:
        super().__init__(client)
        self._value = Value.absent()
        self._synced = asyncio.Event()

    async def _establish_downlink(self) -> None:
        sync_request = _SyncRequest(self.node_uri, self.lane_uri)
        await self.connection._send_message(await sync_request._to_recon())

    async def _receive_event(self, message: '_Envelope') -> None:
        await self.__set_value(message)

    async def _receive_synced(self) -> None:
        self._synced.set()

    async def _send_message(self, message: '_Envelope') -> None:
        """
        Send a message to the remote agent of the downlink.

        :param message:         - Message to send to the remote agent.
        """
        await self.linked.wait()
        await self.connection._send_message(await message._to_recon())

    async def _get_value(self) -> Any:
        """
        Get the value of the downlink after it has been synced.

        :return:                - The current value of the downlink.
        """
        await self._synced.wait()
        return self._value

    async def __set_value(self, message: '_Envelope') -> None:
        """
        Set the value of the the downlink and trigger the `did_set` callback of the downlink subscribers.

        :param message:        - The message from the remote agent.
        :return:
        """
        old_value = self._value
        converter = RecordConverter.get_converter()
        self._value = converter.record_to_object(message._body, self.downlink_manager.registered_classes,
                                                 self.downlink_manager.strict)

        await self.downlink_manager._subscribers_did_set(self._value, old_value)


class _ValueDownlinkView(_DownlinkView):

    def __init__(self, client: 'SwimClient') -> None:
        super().__init__(client)
        self._did_set_callback = None
        self._initialised = asyncio.Event()

    @after_open
    def get(self, wait_sync: bool = False) -> Any:
        """
        Return the value of the downlink.

        :param wait_sync:       - If True, wait for the initial `sync` to be completed before returning.
                                  If False, return immediately.
        :return:                - The value of the Downlink.
        """
        if wait_sync:
            task = self._client._schedule_task(self.__get_value)
            return task.result()
        else:
            return self._value

    @after_open
    def set(self, value: Any, blocking: bool = False) -> None:
        """
        Send a command message to set the value of the lane on the remote agent to the given value.

        :param blocking:        - If True, block until the value has been sent to the server.
        :param value:           - New value for the lane of the remote agent.
        """
        task = self._client._schedule_task(self._send_message, value)

        if blocking:
            task.result()

    def did_set(self, function: Callable) -> '_ValueDownlinkView':
        """
        Set the `did_set` callback of the current downlink view to a given function.

        :param function:   - Function to be called when a value is received by the downlink.
        :return:           - The current downlink view.
        """
        self._did_set_callback = validate_callback(function)
        return self

    @property
    def _value(self) -> 'Any':
        if self._model is None:
            return Value.absent()
        else:
            return self._model._value

    async def _register_manager(self, manager: '_DownlinkManager') -> None:
        await self._assign_manager(manager)

        if manager._is_open:
            await self._execute_did_set(self._model._value, Value.absent())

        self._initialised.set()

    async def _create_downlink_model(self, downlink_manager: '_DownlinkManager') -> '_ValueDownlinkModel':
        model = _ValueDownlinkModel(self._client)
        await self._initalise_model(downlink_manager, model)
        return model

    async def _send_message(self, value: Any) -> None:
        """
        Send a message to the remote agent of the downlink.

        :param value:           - New value for the lane of the remote agent.
        """
        await self._initialised.wait()
        recon = RecordConverter.get_converter().object_to_record(value)
        message = _CommandMessage(self._node_uri, self._lane_uri, recon)

        await self._model._send_message(message)

    # noinspection PyAsyncCall
    async def _execute_did_set(self, current_value: Any, old_value: Any) -> None:
        """
        Execute the custom `did_set` callback of the current downlink view.

        :param current_value:       - The new value of the downlink.
        :param old_value:           - The previous value of the downlink.
        """
        if self._did_set_callback:
            self._client._schedule_task(self._did_set_callback, current_value, old_value)

    async def __get_value(self) -> 'Any':
        await self._initialised.wait()
        return await self._model._get_value()


class _MapDownlinkModel(_DownlinkModel):

    def __init__(self, client: 'SwimClient') -> None:
        super().__init__(client)
        self._map = {}
        self._synced = asyncio.Event()

    async def _establish_downlink(self) -> None:
        sync_request = _SyncRequest(self.node_uri, self.lane_uri)
        await self.connection._send_message(await sync_request._to_recon())

    async def _receive_event(self, message: '_Envelope') -> None:
        if message._body._tag == 'update':
            await self.__receive_update(message)
        if message._body._tag == 'remove':
            await self.__receive_remove(message)

    async def _receive_synced(self) -> None:
        self._synced.set()

    async def _send_message(self, message: '_Envelope') -> None:
        """
        Send a message to the remote agent of the downlink.

        :param message:         - Message to send to the remote agent.
        """
        await self.linked.wait()
        await self.connection._send_message(await message._to_recon())

    async def _get_value(self, key) -> Any:
        """
        Get a value from the map of the downlink using a given key, after it has been synced.

        :param key              - The key of the entry.
        :return:                - The current value of the downlink.
        """
        await self._synced.wait()
        return self._map.get(key, (Value.absent(), Value.absent()))[1]

    async def _get_values(self) -> list:
        """
        Get all of the values from the map of the downlink as a list.

        :return:                - A list with all the values from the downlink map.
        """
        await self._synced.wait()
        return list(self._map.values())

    async def __receive_update(self, message: '_Envelope') -> None:
        key = RecordConverter.get_converter().record_to_object(message._body._get_head().value._get_head().value,
                                                               self.downlink_manager.registered_classes,
                                                               self.downlink_manager.strict)

        value = RecordConverter.get_converter().record_to_object(message._body.get_body(),
                                                                 self.downlink_manager.registered_classes,
                                                                 self.downlink_manager.strict)

        recon_key = await Recon.to_string(message._body._get_head().value._get_head().value)
        old_value = await self._get_value(recon_key)

        self._map[recon_key] = (key, value)
        await self.downlink_manager._subscribers_did_update(key, value, old_value)

    async def __receive_remove(self, message: '_Envelope') -> None:
        key = RecordConverter.get_converter().record_to_object(message._body._get_head().value._get_head().value,
                                                               self.downlink_manager.registered_classes,
                                                               self.downlink_manager.strict)

        recon_key = await Recon.to_string(message._body._get_head().value._get_head().value)
        old_value = self._map.pop(recon_key, (Value.absent(), Value.absent()))[1]

        await self.downlink_manager._subscribers_did_remove(key, old_value)


class _MapDownlinkView(_DownlinkView):

    def __init__(self, client: 'SwimClient') -> None:
        super().__init__(client)
        self._did_update_callback = None
        self._did_remove_callback = None
        self._initialised = asyncio.Event()

    @after_open
    def get(self, key: Any, wait_sync: bool = False) -> Any:
        if wait_sync:
            task = self._client._schedule_task(self.__get_value, key)
            return task.result()
        else:
            return self._map(key)

    @after_open
    def get_all(self, wait_sync: bool = False) -> list:
        if wait_sync:
            task = self._client._schedule_task(self.__get_all_values)
            return task.result()
        else:
            return self._map(None)

    @after_open
    def put(self, key: Any, value: Any, blocking: bool = False) -> None:
        """
        Send a command message to put the given key and value in the remote map lane.

        :param key:             - Entry key.
        :param value:           - Entry value.
        :param blocking:        - If True, block until the value has been sent to the server.
        """
        task = self._client._schedule_task(self.__put_message, key, value)

        if blocking:
            task.result()

    @after_open
    def remove(self, key: Any, blocking: bool = False) -> None:
        """
        Send a command message to remove the given key from the remote map lane.

        :param key:             - Entry key.
        :param blocking:        - If True, block until the value has been sent to the server.
        """
        task = self._client._schedule_task(self.__remove_message, key)

        if blocking:
            task.result()

    def did_update(self, function: Callable) -> '_MapDownlinkView':
        """
        Set the `did_update` callback of the current downlink view to a given function.

        :param function:   - Function to be called when an update event is received by the downlink.
        :return:           - The current downlink view.
        """
        self._did_update_callback = validate_callback(function)
        return self

    def did_remove(self, function: Callable) -> '_MapDownlinkView':
        """
        Set the `did_remove` callback of the current downlink view to a given function.

        :param function:   - Function to be called when a remove event is received by the downlink.
        :return:           - The current downlink view.
        """
        self._did_remove_callback = validate_callback(function)
        return self

    async def _register_manager(self, manager: '_DownlinkManager') -> None:
        await self._assign_manager(manager)

        if manager._is_open:
            for key, value in self._model._map.values():
                await self._execute_did_update(key, value, Value.absent())

        self._initialised.set()

    async def _create_downlink_model(self, downlink_manager: '_DownlinkManager') -> '_MapDownlinkModel':
        model = _MapDownlinkModel(self._client)
        await self._initalise_model(downlink_manager, model)
        return model

    def _map(self, key: Any) -> [Value, dict]:
        if self._model is None:
            return Value.absent()
        else:
            if key is None:
                return list(self._model._map.values())
            else:
                return self._model._map.get(key, (Value.absent(), Value.absent()))[1]

    # noinspection PyAsyncCall
    async def _execute_did_update(self, key: Any, new_value: Any, old_value: Any) -> None:
        """
        Execute the custom `did_update` callback of the current downlink view.

        :param key:             - The entry key of the item.
        :param new_value:       - The new value of the item.
        :param old_value:       - The current value of the item.
        """
        if self._did_update_callback:
            self._client._schedule_task(self._did_update_callback, key, new_value, old_value)

    # noinspection PyAsyncCall
    async def _execute_did_remove(self, key: Any, old_value: Any) -> None:
        """
        Execute the custom `did_remove` callback of the current downlink view.

        :param key:             - The entry key of the item.
        :param old_value:       - The current value of the item.
        """
        if self._did_remove_callback:
            self._client._schedule_task(self._did_remove_callback, key, old_value)

    async def __get_value(self, key: Any) -> Any:
        await self._initialised.wait()
        return await self._model._get_value(key)

    async def __get_all_values(self) -> list:
        await self._initialised.wait()
        return await self._model._get_values()

    async def __put_message(self, key: Any, value: Any) -> None:
        """
        Send a `put` message to the remote agent of the downlink.

        :param key:             - Key for the new entry in the map lane of the remote agent.
        :param value:           - Value for the new entry in the map lane of the remote agent.
        """
        await self._initialised.wait()

        message = _CommandMessage(self._node_uri, self._lane_uri, UpdateRequest(key, value).to_record())
        await self._model._send_message(message)

    async def __remove_message(self, key: Any) -> None:
        """
        Send a `remove` message to the remote agent of the downlink.

        :param key:             - Key for the entry in the map lane that should be removed from the remote agent.
        """
        await self._initialised.wait()

        message = _CommandMessage(self._node_uri, self._lane_uri, RemoveRequest(key).to_record())
        await self._model._send_message(message)



#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import inspect
import sys
from abc import ABC, abstractmethod
from typing import Any, Callable
from swimai.structures import RecordMap, Slot, Text, RecordConverter, Attr
from swimai.structures._structs import _Item, _Record


def before_open(function: 'Callable') -> 'Callable':
    """
    Decorator that allows for a given downlink method to be executed only before the downlink has been opened.
    The downlink should have an `is_open` flag set to False.

    :param function:        - Decorated downlink method.
    :return:                - The return value of the decorated downlink method.
    """

    def wrapper(*args, **kwargs):
        if not args[0]._is_open:
            return function(*args, **kwargs)
        else:
            try:
                raise Exception(f'Cannot execute "{function.__name__}" after the downlink has been opened!')
            except Exception:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                args[0]._client._handle_exception(exc_value, exc_traceback)

    return wrapper


def after_open(function: 'Callable') -> 'Callable':
    """
    Decorator that allows for a given downlink method to be executed only after the downlink has been opened.
    The downlink should have an `is_open` flag set to True.

    :param function:        - Decorated downlink method.
    :return:                - The return value of the decorated downlink method.
    """

    def wrapper(*args, **kwargs):
        if args[0]._is_open:
            return function(*args, **kwargs)
        else:
            try:
                raise Exception(f'Cannot execute "{function.__name__}" before the downlink has been opened!')
            except Exception:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                args[0]._client._handle_exception(exc_value, exc_traceback)

    return wrapper


def convert_to_async(function: 'Any') -> 'Callable':
    """
    Convert regular function into a asynchronous function.

    :param function:        - Function to convert.
    :return:                - Converted asynchronous function.
    """

    async def async_func(*args):
        return function(*args)

    return async_func


def validate_callback(callback: 'Callable') -> 'Callable':
    """
    Validate if a callback is an asynchronous function. If it is a normal function,
    convert it to asynchronous. Otherwise, raise an exception.

    :param callback:        - Callback to validate.
    :return:                - Asynchronous callback.
    """
    if not inspect.iscoroutinefunction(callback) and isinstance(callback, Callable):
        callback = convert_to_async(callback)

    if inspect.iscoroutinefunction(callback):
        return callback
    else:
        raise TypeError('Callback must be a coroutine or a function!')


class MapRequest(ABC):

    def __init__(self, key: Any, value: Any = None) -> None:
        self.key = key
        self.value = value

    @abstractmethod
    def to_record(self):
        """
        Convert the request to a Record object.

        :return:            - Record object representing the request.
        """
        raise NotImplementedError

    def get_key_item(self) -> '_Record':
        """
        Convert the request key into an Item object.

        :return:            - Request key as an Item object.
        """
        key_slot = RecordMap.create()
        key_slot.add(
            Slot.create_slot(Text.create_from('key'), RecordConverter.get_converter().object_to_record(self.key)))

        return key_slot

    def get_value_item(self) -> '_Item':
        """
        Convert the request value into an Item object.

        :return:            - Request value as an Item object.
        """
        value_slot = RecordConverter.get_converter().object_to_record(self.value)
        return value_slot


class UpdateRequest(MapRequest):

    def to_record(self) -> '_Record':
        key_slot = self.get_key_item()
        value_slot = self.get_value_item()

        update_record = RecordMap.create_record_map(Attr.create_attr(Text.create_from('update'), key_slot))
        update_record.add(value_slot)
        return update_record


class RemoveRequest(MapRequest):
    def to_record(self) -> '_Record':
        key_slot = self.get_key_item()

        remove_record = RecordMap.create_record_map(Attr.create_attr(Text.create_from('remove'), key_slot))
        return remove_record



#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import inspect
import unittest
from collections.abc import Callable
from unittest.mock import patch

from aiounittest import async_test

from swimai import SwimClient
from swimai.client._downlinks._downlinks import _EventDownlinkView, _ValueDownlinkView
from swimai.client._downlinks._utils import UpdateRequest, RemoveRequest, convert_to_async, validate_callback
from swimai.structures import RecordMap, Slot, Num, Attr, Value
from test.utils import MockPerson, mock_func, mock_coro


class TestDownlinkUtils(unittest.TestCase):

    def test_before_open_valid_with_args(self):
        # Given
        # noinspection PyTypeChecker
        downlink_view = _EventDownlinkView(None)
        downlink_view._is_open = False
        # When
        downlink_view.set_node_uri('foo')
        # Then
        self.assertEqual('foo', downlink_view._node_uri)

    @patch('warnings.warn')
    def test_before_open_invalid_with_args(self, mock_warn):
        # Given
        # noinspection PyTypeChecker
        downlink_view = _EventDownlinkView(None)
        downlink_view._is_open = True
        downlink_view._client = SwimClient()
        # When
        downlink_view.set_node_uri('foo')
        # Then
        self.assertEqual('Cannot execute "set_node_uri" after the downlink has been opened!',
                         mock_warn.call_args_list[0][0][0])

    def test_before_open_valid_with_kwargs(self):
        # Given
        # noinspection PyTypeChecker
        downlink_view = _EventDownlinkView(None)
        downlink_view._is_open = False
        # When
        downlink_view.set_node_uri(node_uri='foo')
        # Then
        self.assertEqual('foo', downlink_view._node_uri)

    @patch('warnings.warn')
    def test_before_open_invalid_with_kwargs(self, mock_warn):
        # Given
        # noinspection PyTypeChecker
        downlink_view = _EventDownlinkView(None)
        downlink_view._is_open = True
        downlink_view._client = SwimClient()
        # When
        downlink_view.set_node_uri(node_uri='foo')
        # Then
        self.assertEqual('Cannot execute "set_node_uri" after the downlink has been opened!',
                         mock_warn.call_args_list[0][0][0])

    def test_after_open_valid_with_args(self):
        # Given
        # noinspection PyTypeChecker
        downlink_view = _ValueDownlinkView(None)
        downlink_view._is_open = True
        # When
        actual = downlink_view.get(False)
        # Then
        self.assertEqual(Value.absent(), actual)

    @patch('warnings.warn')
    def test_after_open_invalid_with_args(self, mock_warn):
        # Given
        # noinspection PyTypeChecker
        downlink_view = _ValueDownlinkView(None)
        downlink_view._client = SwimClient()
        downlink_view._is_open = False
        # When
        downlink_view.get(False)
        # Then
        self.assertEqual('Cannot execute "get" before the downlink has been opened!', mock_warn.call_args_list[0][0][0])

    def test_after_open_valid_with_kwargs(self):
        # Given
        # noinspection PyTypeChecker
        downlink_view = _ValueDownlinkView(None)
        downlink_view._is_open = True
        # When
        actual = downlink_view.get(wait_sync=False)
        # Then
        self.assertEqual(Value.absent(), actual)

    @patch('warnings.warn')
    def test_after_open_invalid_with_kwargs(self, mock_warn):
        # Given
        # noinspection PyTypeChecker
        downlink_view = _ValueDownlinkView(None)
        downlink_view._is_open = False
        downlink_view._client = SwimClient()
        # When
        downlink_view.get(wait_sync=False)
        # Then
        self.assertEqual('Cannot execute "get" before the downlink has been opened!', mock_warn.call_args_list[0][0][0])

    def test_map_request_get_key_item_primitive(self):
        # Given
        map_request = UpdateRequest('Foo', 23)
        # When
        actual = map_request.get_key_item()
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertIsInstance(actual.get_item(0), Slot)
        self.assertEqual('key', actual.get_item(0).key.value)
        self.assertEqual('Foo', actual.get_item(0).value.value)
        self.assertEqual('Record(Slot("key", "Foo"))', str(actual))

    def test_map_request_get_key_item_complex(self):
        # Given
        mock_person = MockPerson('Bar', 29)
        map_request = UpdateRequest(mock_person, 1)
        # When
        actual = map_request.get_key_item()
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertIsInstance(actual.get_item(0), Slot)
        self.assertEqual('key', actual.get_item(0).key.value)
        self.assertIsInstance(actual.get_item(0).value, RecordMap)
        self.assertIsInstance(actual.get_item(0).value.get_item(0), Attr)
        self.assertEqual('MockPerson', actual.get_item(0).value.get_item(0).key.value)
        self.assertIsInstance(actual.get_item(0).value.get_item(1), Slot)
        self.assertEqual('name', actual.get_item(0).value.get_item(1).key.value)
        self.assertEqual('Bar', actual.get_item(0).value.get_item(1).value.value)
        self.assertIsInstance(actual.get_item(0).value.get_item(2), Slot)
        self.assertEqual('age', actual.get_item(0).value.get_item(2).key.value)
        self.assertEqual(29, actual.get_item(0).value.get_item(2).value.value)
        self.assertEqual('Record(Slot("key", Record(Attr("MockPerson"), Slot("name", "Bar"), Slot("age", 29))))',
                         str(actual))

    def test_map_request_get_value_item_primitive(self):
        # Given
        map_request = UpdateRequest('Foo', 23)
        # When
        actual = map_request.get_value_item()
        # Then
        self.assertIsInstance(actual, Num)
        self.assertEqual(23, actual.value)
        self.assertEqual('23', str(actual))

    def test_map_request_get_value_item_complex(self):
        # Given
        mock_person = MockPerson('Foo', 20)
        map_request = UpdateRequest('Foo', mock_person)
        # When
        actual = map_request.get_value_item()
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertIsInstance(actual.get_item(0), Attr)
        self.assertEqual('MockPerson', actual.get_item(0).key.value)
        self.assertIsInstance(actual.get_item(1), Slot)
        self.assertEqual('name', actual.get_item(1).key.value)
        self.assertEqual('Foo', actual.get_item(1).value.value)
        self.assertIsInstance(actual.get_item(2), Slot)
        self.assertEqual('age', actual.get_item(2).key.value)
        self.assertEqual(20, actual.get_item(2).value.value)
        self.assertEqual('Record(Attr("MockPerson"), Slot("name", "Foo"), Slot("age", 20))', str(actual))

    def test_update_request_to_record(self):
        # Given
        update_request = UpdateRequest('Moo', 21)
        # When
        actual = update_request.to_record()
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertIsInstance(actual.get_item(0), Attr)
        self.assertEqual('update', actual.get_item(0).key.value)
        self.assertIsInstance(actual.get_item(0).value, RecordMap)
        self.assertIsInstance(actual.get_item(0).value.get_item(0), Slot)
        self.assertEqual('key', actual.get_item(0).value.get_item(0).key.value)
        self.assertEqual('Moo', actual.get_item(0).value.get_item(0).value.value)
        self.assertIsInstance(actual.get_item(1), Num)
        self.assertEqual(21, actual.get_item(1).value)
        self.assertEqual('Record(Attr("update", Record(Slot("key", "Moo"))), 21)', str(actual))

    def test_remove_request_to_record(self):
        # Given
        remove_request = RemoveRequest('Boo', 99)
        # When
        actual = remove_request.to_record()
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertIsInstance(actual.get_item(0), Attr)
        self.assertEqual('remove', actual.get_item(0).key.value)
        self.assertIsInstance(actual.get_item(0).value, RecordMap)
        self.assertIsInstance(actual.get_item(0).value.get_item(0), Slot)
        self.assertEqual('key', actual.get_item(0).value.get_item(0).key.value)
        self.assertEqual('Boo', actual.get_item(0).value.get_item(0).value.value)
        self.assertEqual('Record(Attr("remove", Record(Slot("key", "Boo"))))', str(actual))

    @async_test
    async def test_convert_to_async_function(self):
        # Given
        function = mock_func
        # When
        actual = convert_to_async(function)
        response = await actual()
        # Then
        self.assertIsInstance(actual, Callable)
        self.assertTrue(inspect.iscoroutinefunction(actual))
        self.assertEqual('mock_func_response', response)

    @async_test
    async def test_downlink_view_validate_callback_function(self):
        # Given
        func = mock_func
        # When
        actual = validate_callback(func)
        result = await actual()
        # Then
        self.assertEqual('mock_func_response', result)
        self.assertTrue(isinstance(actual, Callable))
        self.assertTrue(inspect.iscoroutinefunction(actual))

    @async_test
    async def test_downlink_view_validate_coro(self):
        # Given
        coro = mock_coro
        # When
        actual = validate_callback(coro)
        result = await actual()
        # Then
        self.assertEqual('mock_coro_response', result)
        self.assertTrue(isinstance(actual, Callable))
        self.assertTrue(inspect.iscoroutinefunction(actual))

    @async_test
    async def test_downlink_view_validate_invalid(self):
        # Given
        integer = 31
        # When
        with self.assertRaises(TypeError) as error:
            # noinspection PyTypeChecker
            validate_callback(integer)

        # Then
        message = error.exception.args[0]
        self.assertEqual(message, 'Callback must be a coroutine or a function!')



#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import math
from abc import ABC, abstractmethod
from typing import Optional
from swimai.recon import Recon
from swimai.structures import Attr, Value, Num, RecordMap
from swimai.structures._structs import _Record, _Item


class _Envelope(ABC):

    def __init__(self, node_uri: str, lane_uri: str, tag: str, form: '_Form', body: _Item = Value.absent()) -> None:
        self._node_uri = node_uri
        self._lane_uri = lane_uri
        self._tag = tag
        self._form = form
        self._body = body

    @staticmethod
    def _create_from_value(value: RecordMap) -> '_Envelope':
        """
        Parse a Swim value object into an Envelope.

        :param value:           - Swim value object.
        :return:                - Envelope from the Swim value object.
        """
        tag = value._tag
        form = _Envelope._resolve_form(tag)

        return form._cast(value)

    def _to_value(self) -> Value:
        """
        Create a Swim value object representing this Envelope.

        :return:                - Swim value object from this Envelope.
        """
        return self._form._mold(self)

    @property
    def _route(self) -> str:
        return f'{self._node_uri}/{self._lane_uri}'

    @staticmethod
    async def _parse_recon(recon_message: str) -> Optional['_Envelope']:
        """
        Parse a Recon message in string format into an Envelope.

        :param recon_message    - Recon message in string format.
        :return:                - Envelope from the Recon message.
        """
        value = await Recon.parse(recon_message)
        if isinstance(value, RecordMap):
            return _Envelope._create_from_value(value)

    @staticmethod
    def _resolve_form(tag: str) -> '_Form':
        """
        Returns a Swim form corresponding to a given tag.

        :param tag:             - Name of the tag as string.
        :return:                - The form corresponding to the tag.
        """

        if tag == 'link':
            return _LinkRequestForm()
        if tag == 'sync':
            return _SyncRequestForm()
        if tag == 'synced':
            return _SyncedResponseForm()
        if tag == 'linked':
            return _LinkedResponseForm()
        if tag == 'unlinked':
            return _UnlinkedResponseForm()
        if tag == 'event':
            return _EventMessageForm()
        if tag == 'command':
            return _CommandMessageForm()
        else:
            raise TypeError(f'Invalid form tag: {tag}')

    async def _to_recon(self) -> str:
        """
        Create a Recon message in string format representing this Envelope.

        :return:                - Recon message in string format from this Envelope.
        """
        return await Recon.to_string(self._to_value())


class _LinkAddressedEnvelope(_Envelope):

    def __init__(self, node_uri: str, lane_uri: str, prio: float, rate: float, tag: str, form: '_Form',
                 body: _Item = Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, tag, form, body)
        self._prio = prio
        self._rate = rate


class _LaneAddressedEnvelope(_Envelope):

    def __init__(self, node_uri: str, lane_uri: str, tag: str, form: '_Form', body=Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, tag, form, body)


class _LinkRequest(_LinkAddressedEnvelope):

    def __init__(self, node_uri: str, lane_uri: str, prio: float = 0.0, rate: float = 0.0,
                 body: _Item = Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, prio, rate, tag='link', form=_LinkRequestForm(), body=body)


class _SyncRequest(_LinkAddressedEnvelope):

    def __init__(self, node_uri: str, lane_uri: str, prio: float = 0.0, rate: float = 0.0,
                 body: _Item = Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, prio, rate, tag='sync', form=_SyncRequestForm(), body=body)


class _LinkedResponse(_LinkAddressedEnvelope):

    def __init__(self, node_uri: str, lane_uri: str, prio: float = 0.0, rate: float = 0.0,
                 body: _Item = Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, prio, rate, tag='linked', form=_LinkedResponseForm(), body=body)


class _UnlinkedResponse(_LinkAddressedEnvelope):
    def __init__(self, node_uri: str, lane_uri: str, prio: float = 0.0, rate: float = 0.0,
                 body: _Item = Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, prio, rate, tag='unlinked', form=_UnlinkedResponseForm(), body=body)


class _SyncedResponse(_LaneAddressedEnvelope):

    def __init__(self, node_uri: str, lane_uri: str, body: _Item = Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, tag='synced', form=_SyncedResponseForm(), body=body)


class _CommandMessage(_LaneAddressedEnvelope):

    def __init__(self, node_uri: str, lane_uri: str, body: _Item = Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, tag='command', form=_CommandMessageForm(), body=body)


class _EventMessage(_LaneAddressedEnvelope):

    def __init__(self, node_uri: str, lane_uri: str, body: _Item = Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, tag='event', form=_EventMessageForm(), body=body)


class _Form(ABC):

    @property
    @abstractmethod
    def _tag(self) -> str:
        """
        Return the tag associated with the given Form object.

        :return:                - Name of the tag as string value.
        """
        raise NotImplementedError

    @abstractmethod
    def _mold(self, envelope: '_Envelope') -> 'Value':
        """
        Create a value object from a given Envelope.

        :param envelope:        - Envelope to convert to Value object.
        :return:                - Value object representing the Envelope.
        """
        raise NotImplementedError

    @abstractmethod
    def _cast(self, item: RecordMap) -> '_Envelope':
        """
        Create an Envelope object from a RecordMap.

        :param item:            - RecordMap to convert to Envelope.
        :return:                - Envelope object created from the RecordMap.
        """
        raise NotImplementedError


class _LinkAddressedForm(_Form):

    @abstractmethod
    def _create_envelope_from(self, node_uri: str, lane_uri: str, prio: float, rate: float,
                              body: '_Item') -> '_Envelope':
        """
        Create an Envelope object corresponding to the given LinkAddressedForm.

        :param node_uri:        - Node URI of the Envelope.
        :param lane_uri:        - Lane URI of the Envelope.
        :param prio:            - Priority of the Envelope.
        :param rate:            - Rate of the Envelope.
        :param body:            - Body of the Envelope.
        :return:                - Envelope corresponding to the LinkAddressedForm.
        """
        raise NotImplementedError

    def _mold(self, envelope: Optional['_LinkAddressedEnvelope']) -> 'Value':

        if envelope is not None:

            headers = _Record.create()._add_slot('node', envelope._node_uri)._add_slot('lane', envelope._lane_uri)
            prio = envelope._prio

            if prio != 0 and not math.isnan(prio):
                headers._add_slot('prio', Num.create_from(prio))

            rate = envelope._rate

            if rate != 0 and not math.isnan(rate):
                headers._add_slot('rate', Num.create_from(rate))

            return Attr.create_attr(self._tag, headers)._concat(envelope._body)
        else:
            return _Item.extant()

    def _cast(self, value: RecordMap) -> Optional['_Envelope']:
        headers = value._get_headers(self._tag)
        node_uri = None
        lane_uri = None
        prio = 0.0
        rate = 0.0

        for index in range(0, headers.size):
            header = headers.get_item(index)
            key = header.key.get_string_value()

            if key is not None:

                if key == 'node':
                    node_uri = header.value.get_string_value()
                elif key == 'lane':
                    lane_uri = header.value.get_string_value()
                elif key == 'prio':
                    prio = header.value.get_num_value()
                elif key == 'rate':
                    rate = header.value.get_num_value()

        if node_uri is not None and lane_uri is not None:
            body = value.get_body()
            return self._create_envelope_from(node_uri, lane_uri, prio, rate, body)


class _LaneAddressedForm(_Form):

    @abstractmethod
    def _create_envelope_from(self, node_uri: str, lane_uri: str, body: _Item) -> '_Envelope':
        """
        :param node_uri:        - Node URI of the Envelope.
        :param lane_uri:        - Lane URI of the Envelope.
        :param body:            - Body of the Envelope.
        :return:                - Envelope corresponding to the LaneAddressedForm.
        """
        raise NotImplementedError

    def _mold(self, envelope: Optional['_LaneAddressedEnvelope']) -> 'Value':

        if envelope is not None:
            headers = _Record.create()._add_slot('node', envelope._node_uri)._add_slot('lane', envelope._lane_uri)
            return Attr.create_attr(self._tag, headers)._concat(envelope._body)
        else:
            return _Item.extant()

    def _cast(self, item: 'RecordMap') -> Optional['_Envelope']:
        value = item
        headers = value._get_headers(self._tag)
        node_uri = None
        lane_uri = None

        for index in range(0, headers.size):
            header = headers.get_item(index)
            key = header.key.get_string_value()

            if key is not None:

                if key == 'node':
                    node_uri = header.value.get_string_value()
                elif key == 'lane':
                    lane_uri = header.value.get_string_value()

        if node_uri is not None and lane_uri is not None:
            body = value.get_body()
            return self._create_envelope_from(node_uri, lane_uri, body)


class _SyncRequestForm(_LinkAddressedForm):

    @property
    def _tag(self) -> str:
        return 'sync'

    def _create_envelope_from(self, node_uri: str, lane_uri: str, prio: float, rate: float, body: _Item) -> '_Envelope':
        return _SyncRequest(node_uri, lane_uri, prio, rate, body=body)


class _SyncedResponseForm(_LaneAddressedForm):

    @property
    def _tag(self) -> str:
        return 'synced'

    def _create_envelope_from(self, node_uri: str, lane_uri: str, body: _Item) -> '_Envelope':
        return _SyncedResponse(node_uri, lane_uri, body=body)


class _LinkRequestForm(_LinkAddressedForm):

    @property
    def _tag(self) -> str:
        return 'link'

    def _create_envelope_from(self, node_uri: str, lane_uri: str, prio: float, rate: float, body: _Item) -> '_Envelope':
        return _LinkRequest(node_uri, lane_uri, prio, rate, body=body)


class _LinkedResponseForm(_LinkAddressedForm):

    @property
    def _tag(self) -> str:
        return 'linked'

    def _create_envelope_from(self, node_uri: str, lane_uri: str, prio: float, rate: float, body: _Item) -> '_Envelope':
        return _LinkedResponse(node_uri, lane_uri, prio, rate, body=body)


class _UnlinkedResponseForm(_LinkAddressedForm):

    @property
    def _tag(self) -> str:
        return 'unlinked'

    def _create_envelope_from(self, node_uri: str, lane_uri: str, prio: float, rate: float,
                              body: '_Item') -> '_Envelope':
        return _UnlinkedResponse(node_uri, lane_uri, prio, rate, body=body)


class _CommandMessageForm(_LaneAddressedForm):

    @property
    def _tag(self) -> str:
        return 'command'

    def _create_envelope_from(self, node_uri: str, lane_uri: str, body: _Item) -> '_Envelope':
        return _CommandMessage(node_uri, lane_uri, body=body)


class _EventMessageForm(_LaneAddressedForm):

    @property
    def _tag(self) -> str:
        return 'event'

    def _create_envelope_from(self, node_uri: str, lane_uri: str, body: _Item) -> '_Envelope':
        return _EventMessage(node_uri, lane_uri, body)



#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

from aiounittest import async_test
from swimai.warp._warp import _Envelope


class TestParser(unittest.TestCase):

    @async_test
    async def test_parse_sync(self):
        # Given
        message = '@sync(node: foo, lane: bar)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('sync', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)

    @async_test
    async def test_parse_sync_escaped(self):
        # Given
        message = '@sync(node: "bar/baz/2", lane: "foo/bar")'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('sync', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)

    @async_test
    async def test_parse_sync_body_int(self):
        # Given
        message = '@sync(node: "bar/baz/2", lane: "foo/bar")43'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('sync', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual(43, actual._body.value)

    @async_test
    async def test_parse_sync_body_float(self):
        # Given
        message = '@sync(node: "bar/baz/2", lane: "foo/bar")0.31'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('sync', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual(0.31, actual._body.value)

    @async_test
    async def test_parse_sync_body_bool(self):
        # Given
        message = '@sync(node: "bar/baz/2", lane: "foo/bar")false'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('sync', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual(False, actual._body.value)

    @async_test
    async def test_parse_sync_body_string(self):
        # Given
        message = '@sync(node: "bar/baz/2", lane: "foo/bar")"Hello, World"'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('sync', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual('Hello, World', actual._body.value)

    @async_test
    async def test_parse_sync_prio(self):
        # Given
        message = '@sync(node: foo, lane: bar, prio: 3.2)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('sync', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual(3.2, actual._prio)

    @async_test
    async def test_parse_sync_rate(self):
        # Given
        message = '@sync(node: foo, lane: bar, rate: 33)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('sync', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual(33, actual._rate)

    @async_test
    async def test_parse_sync_prio_rate(self):
        # Given
        message = '@sync(node: foo, lane: bar, prio: 13, rate: 37)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('sync', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual(13, actual._prio)
        self.assertEqual(37, actual._rate)

    @async_test
    async def test_parse_synced(self):
        # Given
        message = '@synced(node: foo, lane: bar)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('synced', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)

    @async_test
    async def test_parse_synced_escaped(self):
        # Given
        message = '@synced(node: "foo/bar", lane: "lane/uri/test")'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('synced', actual._tag)
        self.assertEqual('foo/bar', actual._node_uri)
        self.assertEqual('lane/uri/test', actual._lane_uri)

    @async_test
    async def test_parse_synced_body_int(self):
        # Given
        message = '@synced(node: foo, lane: bar)33'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('synced', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual(33, actual._body.value)

    @async_test
    async def test_parse_synced_body_float(self):
        # Given
        message = '@synced(node: foo, lane: bar)37.13'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('synced', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual(37.13, actual._body.value)

    @async_test
    async def test_parse_synced_body_bool(self):
        # Given
        message = '@synced(node: foo, lane: bar)true'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('synced', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual(True, actual._body.value)

    @async_test
    async def test_parse_synced_body_string(self):
        # Given
        message = '@synced(node: foo, lane: bar)"Hello, World"'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('synced', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual('Hello, World', actual._body.value)

    @async_test
    async def test_parse_link(self):
        # Given
        message = '@link(node: boo, lane: far)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('link', actual._tag)
        self.assertEqual('boo', actual._node_uri)
        self.assertEqual('far', actual._lane_uri)

    @async_test
    async def test_parse_link_escaped(self):
        # Given
        message = '@link(node: "bar/baz/5", lane: "foo/baz")'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('link', actual._tag)
        self.assertEqual('bar/baz/5', actual._node_uri)
        self.assertEqual('foo/baz', actual._lane_uri)

    @async_test
    async def test_parse_link_body_int(self):
        # Given
        message = '@link(node: "baz/bar/2", lane: "foo/bar")91'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('link', actual._tag)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual('baz/bar/2', actual._node_uri)
        self.assertEqual(91, actual._body.value)

    @async_test
    async def test_parse_link_body_float(self):
        # Given
        message = '@link(node: "bar/baz/2", lane: "foo/bar")-0.3'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('link', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual(-0.3, actual._body.value)

    @async_test
    async def test_parse_link_body_bool(self):
        # Given
        message = '@link(node: "bar/baz/2", lane: "foo/bar")false'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('link', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual(False, actual._body.value)

    @async_test
    async def test_parse_link_body_string(self):
        # Given
        message = '@link(node: "bar/baz/2", lane: "foo/bar")"Hello, World"'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('link', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual('Hello, World', actual._body.value)

    @async_test
    async def test_parse_link_prio(self):
        # Given
        message = '@link(node: foo, lane: bar, prio: 1000.3)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('link', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual(1000.3, actual._prio)

    @async_test
    async def test_parse_link_rate(self):
        # Given
        message = '@link(node: "foo", lane: "1/bar/", rate: 33)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('link', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('1/bar/', actual._lane_uri)
        self.assertEqual(33, actual._rate)

    @async_test
    async def test_parse_link_prio_rate(self):
        # Given
        message = '@link(node: foo, lane: bar, prio: 13, rate: 37)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('link', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual(13, actual._prio)
        self.assertEqual(37, actual._rate)

    @async_test
    async def test_parse_linked(self):
        # Given
        message = '@linked(node: foo, lane: bar)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('linked', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)

    @async_test
    async def test_parse_linked_escaped(self):
        # Given
        message = '@linked(node: "bar/baz/2", lane: "foo/bar")'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('linked', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)

    @async_test
    async def test_parse_linked_body_int(self):
        # Given
        message = '@linked(node: "bar/baz/2", lane: "foo/bar")9999999'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('linked', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual(9999999, actual._body.value)

    @async_test
    async def test_parse_linked_body_float(self):
        # Given
        message = '@linked(node: "bar/baz/2", lane: "foo/bar")-0.00031'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('linked', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual(-0.00031, actual._body.value)

    @async_test
    async def test_parse_linked_body_bool(self):
        # Given
        message = '@linked(node: "bar/baz/2", lane: "foo/bar")false'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('linked', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual(False, actual._body.value)

    @async_test
    async def test_parse_linked_body_string(self):
        # Given
        message = '@linked(node: "bar/baz/2", lane: "foo/bar")"Hello, World"'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('linked', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual('Hello, World', actual._body.value)

    @async_test
    async def test_parse_linked_prio(self):
        # Given
        message = '@linked(node: foo, lane: bar, prio: 1000.3)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('linked', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual(1000.3, actual._prio)

    @async_test
    async def test_parse_linked_rate(self):
        # Given
        message = '@linked(node: "foo", lane: "1/bar/", rate: 33)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('linked', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('1/bar/', actual._lane_uri)
        self.assertEqual(33, actual._rate)

    @async_test
    async def test_parse_linked_prio_rate(self):
        # Given
        message = '@linked(node: foo, lane: bar, prio: 13, rate: 37)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('linked', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual(13, actual._prio)
        self.assertEqual(37, actual._rate)

    @async_test
    async def test_parse_unlinked(self):
        # Given
        message = '@unlinked(node: foo, lane: bar)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('unlinked', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)

    @async_test
    async def test_parse_unlinked_escaped(self):
        # Given
        message = '@unlinked(node: "bar/baz/2", lane: "foo/bar")'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('unlinked', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)

    @async_test
    async def test_parse_unlinked_body_int(self):
        # Given
        message = '@unlinked(node: "bar/baz/2", lane: "foo/bar")9999999'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('unlinked', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual(9999999, actual._body.value)

    @async_test
    async def test_parse_unlinked_body_float(self):
        # Given
        message = '@unlinked(node: "bar/baz/2", lane: "foo/bar")-0.00031'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('unlinked', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual(-0.00031, actual._body.value)

    @async_test
    async def test_parse_unlinked_body_bool(self):
        # Given
        message = '@unlinked(node: "bar/baz/2", lane: "foo/bar")false'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('unlinked', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual(False, actual._body.value)

    @async_test
    async def test_parse_unlinked_body_string(self):
        # Given
        message = '@unlinked(node: "bar/baz/2", lane: "foo/bar")"Hello, World"'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('unlinked', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual('Hello, World', actual._body.value)

    @async_test
    async def test_parse_unlinked_prio(self):
        # Given
        message = '@unlinked(node: foo, lane: bar, prio: 1000.3)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual(1000.3, actual._prio)

    @async_test
    async def test_parse_unlinked_rate(self):
        # Given
        message = '@unlinked(node: "foo", lane: "1/bar/", rate: 33)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('1/bar/', actual._lane_uri)
        self.assertEqual(33, actual._rate)

    @async_test
    async def test_parse_unlinked_prio_rate(self):
        # Given
        message = '@unlinked(node: foo, lane: bar, prio: 13, rate: 37)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual(13, actual._prio)
        self.assertEqual(37, actual._rate)

    @async_test
    async def test_parse_command(self):
        # Given
        message = '@command(node: foo, lane: bar)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('command', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)

    @async_test
    async def test_parse_command_escaped(self):
        # Given
        message = '@command(node: "foo/bar", lane: "lane/uri/test")'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('command', actual._tag)
        self.assertEqual('foo/bar', actual._node_uri)
        self.assertEqual('lane/uri/test', actual._lane_uri)

    @async_test
    async def test_parse_command_body_int(self):
        # Given
        message = '@command(node: "foo/bar", lane: "lane/uri/test")-24'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('command', actual._tag)
        self.assertEqual('foo/bar', actual._node_uri)
        self.assertEqual('lane/uri/test', actual._lane_uri)
        self.assertEqual(-24, actual._body.value)

    @async_test
    async def test_parse_command_body_float(self):
        # Given
        message = '@command(node: "foo/bar", lane: "lane/uri/test")-0.5'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('command', actual._tag)
        self.assertEqual('foo/bar', actual._node_uri)
        self.assertEqual('lane/uri/test', actual._lane_uri)
        self.assertEqual(-0.5, actual._body.value)

    @async_test
    async def test_parse_command_body_bool(self):
        # Given
        message = '@command(node: "foo/bar", lane: "lane/uri/test")true'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('command', actual._tag)
        self.assertEqual('foo/bar', actual._node_uri)
        self.assertEqual('lane/uri/test', actual._lane_uri)
        self.assertEqual(True, actual._body.value)

    @async_test
    async def test_parse_command_body_string(self):
        # Given
        message = '@command(node: "foo/bar", lane: "lane/uri/test")"0.32"'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('command', actual._tag)
        self.assertEqual('foo/bar', actual._node_uri)
        self.assertEqual('lane/uri/test', actual._lane_uri)
        self.assertEqual('0.32', actual._body.value)

    @async_test
    async def test_parse_command_body_remove(self):
        # Given
        message = '@command(node:"/unit/foo",lane:shoppingCart)@remove(key:FromClientLink)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('command', actual._tag)
        self.assertEqual('/unit/foo', actual._node_uri)
        self.assertEqual('shoppingCart', actual._lane_uri)
        self.assertEqual('remove', actual._body.get_items()[0].key.value)
        self.assertEqual('key', actual._body.get_items()[0].value.get_items()[0].key.value)
        self.assertEqual('FromClientLink', actual._body.get_items()[0].value.get_items()[0].value.value)

    @async_test
    async def test_parse_event(self):
        # Given
        message = '@event(node: foo, lane: bar)'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('event', actual._tag)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)

    @async_test
    async def test_parse_event_escaped(self):
        # Given
        message = '@event(node: "bar/baz/2", lane: "foo/bar")'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('event', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)

    @async_test
    async def test_parse_event_body_int(self):
        # Given
        message = '@event(node: "bar/baz/2", lane: "foo/bar")332'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('event', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual(332, actual._body.value)

    @async_test
    async def test_parse_event_body_float(self):
        # Given
        message = '@event(node: "bar/baz/2", lane: "foo/bar")0.1'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('event', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual(0.1, actual._body.value)

    @async_test
    async def test_parse_event_body_bool(self):
        # Given
        message = '@event(node: "bar/baz/2", lane: "foo/bar")true'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('event', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual(True, actual._body.value)

    @async_test
    async def test_parse_event_body_string(self):
        # Given
        message = '@event(node: "bar/baz/2", lane: "foo/bar")"Hello, World"'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('event', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)
        self.assertEqual('Hello, World', actual._body.value)

    @async_test
    async def test_parse_event_body_object(self):
        # Given
        message = '@event(node: "bar/baz/2", lane: "foo/bar")@Person{name:Bar,age:14,salary:5.9}'
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('event', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)

        self.assertEqual(4, len(actual._body.get_items()))
        self.assertEqual('Person', actual._body.get_items()[0].key.value)
        self.assertEqual('name', actual._body.get_items()[1].key.value)
        self.assertEqual('Bar', actual._body.get_items()[1].value.value)
        self.assertEqual('age', actual._body.get_items()[2].key.value)
        self.assertEqual(14, actual._body.get_items()[2].value.value)
        self.assertEqual('salary', actual._body.get_items()[3].key.value)
        self.assertEqual(5.9, actual._body.get_items()[3].value.value)

    @async_test
    async def test_parse_event_body_nested(self):
        # Given
        person = '@Person{name:Par,age:11,salary:5.9,friend:@Person{name:Foo,age:18,salary:99.9}'
        message = '@event(node: "bar/baz/2", lane: "foo/bar")' + person
        # When
        actual = await _Envelope._parse_recon(message)
        # Then
        self.assertEqual('event', actual._tag)
        self.assertEqual('bar/baz/2', actual._node_uri)
        self.assertEqual('foo/bar', actual._lane_uri)

        self.assertEqual(5, len(actual._body.get_items()))
        self.assertEqual('Person', actual._body.get_items()[0].key.value)
        self.assertEqual('name', actual._body.get_items()[1].key.value)
        self.assertEqual('Par', actual._body.get_items()[1].value.value)
        self.assertEqual('age', actual._body.get_items()[2].key.value)
        self.assertEqual(11, actual._body.get_items()[2].value.value)
        self.assertEqual('salary', actual._body.get_items()[3].key.value)
        self.assertEqual(5.9, actual._body.get_items()[3].value.value)
        self.assertEqual('friend', actual._body.get_items()[4].key.value)
        self.assertEqual(4, len(actual._body.get_items()[4].value.get_items()))
        self.assertEqual('Person', actual._body.get_items()[4].value.get_items()[0].key.value)
        self.assertEqual('name', actual._body.get_items()[4].value.get_items()[1].key.value)
        self.assertEqual('Foo', actual._body.get_items()[4].value.get_items()[1].value.value)
        self.assertEqual('age', actual._body.get_items()[4].value.get_items()[2].key.value)
        self.assertEqual(18, actual._body.get_items()[4].value.get_items()[2].value.value)
        self.assertEqual('salary', actual._body.get_items()[4].value.get_items()[3].key.value)
        self.assertEqual(99.9, actual._body.get_items()[4].value.get_items()[3].value.value)



#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import math
import unittest

from swimai.structures import Value, Text, RecordMap, Slot, Attr, Num
from swimai.warp._warp import _SyncedResponseForm, _SyncedResponse, _SyncRequestForm, _SyncRequest, _LinkedResponseForm, \
    _CommandMessageForm, _EventMessageForm, _LinkedResponse, _CommandMessage, _EventMessage, _LinkRequestForm, _LinkRequest, \
    _UnlinkedResponseForm, _UnlinkedResponse


class TestForms(unittest.TestCase):

    def test_lane_addressed_form_mold_empty(self):
        # Given
        envelope = None
        form = _SyncedResponseForm()
        # When
        actual = form._mold(envelope)
        # Then
        self.assertEqual(Value.extant(), actual)

    def test_lane_addressed_form_cast_missing_headers(self):
        # Given
        form = _SyncedResponseForm()
        items = RecordMap.create()
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr(Text.create_from('synced'), items))
        # When
        actual = form._cast(record_map)
        # Then
        self.assertEqual(None, actual)

    def test_lane_addressed_form_cast_missing_node(self):
        # Given
        form = _SyncedResponseForm()
        items = RecordMap.create()
        items.add(Attr.create_attr(Text.create_from('lane'), Text.create_from('foo_lane')))
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr(Text.create_from('synced'), items))
        # When
        actual = form._cast(record_map)
        # Then
        self.assertEqual(None, actual)

    def test_lane_addressed_form_cast_missing_lane(self):
        # Given
        form = _SyncedResponseForm()
        items = RecordMap.create()
        items.add(Attr.create_attr(Text.create_from('node'), Text.create_from('foo_node')))
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr(Text.create_from('synced'), items))
        # When
        actual = form._cast(record_map)
        # Then
        self.assertEqual(None, actual)

    def test_link_addressed_form_mold_empty(self):
        # Given
        envelope = None
        form = _SyncRequestForm()
        # When
        actual = form._mold(envelope)
        # Then
        self.assertEqual(Value.extant(), actual)

    def test_link_addressed_form_mold_prio_zero(self):
        # Given
        envelope = _SyncRequest('node_foo', 'lane_bar', prio=0, rate=2.5)
        form = _SyncRequestForm()
        # When
        actual = form._mold(envelope)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual('node_foo', actual.get_item(0).value.get_item(0).value.value)
        self.assertEqual('lane_bar', actual.get_item(0).value.get_item(1).value.value)
        self.assertEqual('rate', actual.get_item(0).value.get_item(2).key.value)
        self.assertEqual(2.5, actual.get_item(0).value.get_item(2).value.value)

    def test_link_addressed_form_mold_prio_nan(self):
        # Given
        envelope = _SyncRequest('synced_node', 'synced_lane', prio=math.nan, rate=3.4)
        form = _SyncRequestForm()
        # When
        actual = form._mold(envelope)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual('synced_node', actual.get_item(0).value.get_item(0).value.value)
        self.assertEqual('synced_lane', actual.get_item(0).value.get_item(1).value.value)
        self.assertEqual('rate', actual.get_item(0).value.get_item(2).key.value)
        self.assertEqual(3.4, actual.get_item(0).value.get_item(2).value.value)

    def test_link_addressed_form_mold_rate_zero(self):
        # Given
        envelope = _SyncRequest('node_rate_foo', 'lane_rate_bar', prio=2.1, rate=0)
        form = _SyncRequestForm()
        # When
        actual = form._mold(envelope)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual('node_rate_foo', actual.get_item(0).value.get_item(0).value.value)
        self.assertEqual('lane_rate_bar', actual.get_item(0).value.get_item(1).value.value)
        self.assertEqual('prio', actual.get_item(0).value.get_item(2).key.value)
        self.assertEqual(2.1, actual.get_item(0).value.get_item(2).value.value)

    def test_link_addressed_form_mold_rate_nan(self):
        # Given
        envelope = _SyncRequest('node_baz', 'lane_qux', prio=13.4, rate=math.nan)
        form = _SyncRequestForm()
        # When
        actual = form._mold(envelope)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual('node_baz', actual.get_item(0).value.get_item(0).value.value)
        self.assertEqual('lane_qux', actual.get_item(0).value.get_item(1).value.value)
        self.assertEqual('prio', actual.get_item(0).value.get_item(2).key.value)
        self.assertEqual(13.4, actual.get_item(0).value.get_item(2).value.value)

    def test_link_addressed_cast_missing_headers(self):
        # Given
        form = _SyncRequestForm()
        items = RecordMap.create()
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr(Text.create_from('sync'), items))
        # When
        actual = form._cast(record_map)
        # Then
        self.assertEqual(None, actual)

    def test_link_addressed_form_cast_missing_node(self):
        # Given
        form = _SyncRequestForm()
        items = RecordMap.create()
        items.add(Attr.create_attr(Text.create_from('lane'), Text.create_from('sync_foo_lane')))
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr(Text.create_from('sync'), items))
        # When
        actual = form._cast(record_map)
        # Then
        self.assertEqual(None, actual)

    def test_link_addressed_form_cast_missing_lane(self):
        # Given
        form = _SyncRequestForm()
        items = RecordMap.create()
        items.add(Attr.create_attr(Text.create_from('node'), Text.create_from('sync_foo_node')))
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr(Text.create_from('sync'), items))
        # When
        actual = form._cast(record_map)
        # Then
        self.assertEqual(None, actual)

    def test_synced_form(self):
        # Given
        actual = _SyncedResponseForm()
        # Then
        self.assertIsInstance(actual, _SyncedResponseForm)
        self.assertEqual('synced', actual._tag)

    def test_synced_form_create_envelope(self):
        # Given
        form = _SyncedResponseForm()
        body = Text.create_from('synced_body')
        # When
        actual = form._create_envelope_from('foo', 'bar', body)
        # Then
        self.assertIsInstance(actual, _SyncedResponse)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual(body, actual._body)

    def test_synced_form_mold(self):
        # Given
        form = _SyncedResponseForm()
        envelope = _SyncedResponse('foo_node', 'bar_lane', body=Text.create_from('Boo'))
        # When
        actual = form._mold(envelope)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(2, actual.size)
        self.assertEqual('synced', actual._tag)
        self.assertEqual('foo_node', actual.get_item(0).value.get_item(0).value.value)
        self.assertEqual('bar_lane', actual.get_item(0).value.get_item(1).value.value)
        self.assertEqual('Boo', actual.get_item(1).value)

    def test_synced_form_cast(self):
        # Given
        form = _SyncedResponseForm()
        items = RecordMap.create()
        items.add(Slot.create_slot(Text.create_from('node'), Text.create_from('synced_node')))
        items.add(Slot.create_slot(Text.create_from('lane'), Text.create_from('synced_lane')))
        body = Attr.create_attr(Text.create_from('body'), Text.create_from('synced_body'))
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr(Text.create_from('synced'), items))
        record_map.add(body)
        # When
        actual = form._cast(record_map)
        # Then
        self.assertIsInstance(actual, _SyncedResponse)
        self.assertEqual('synced_node', actual._node_uri)
        self.assertEqual('synced_lane', actual._lane_uri)
        self.assertEqual(body, actual._body.get_item(0))

    def test_sync_form(self):
        # Given
        actual = _SyncRequestForm()
        # Then
        self.assertIsInstance(actual, _SyncRequestForm)
        self.assertEqual('sync', actual._tag)

    def test_sync_form_create_envelope(self):
        # Given
        form = _SyncRequestForm()
        body = Text.create_from('sync_body')
        # When
        actual = form._create_envelope_from('foo', 'bar', prio=3.2, rate=2.1, body=body)
        # Then
        self.assertIsInstance(actual, _SyncRequest)
        self.assertEqual('foo', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual(3.2, actual._prio)
        self.assertEqual(2.1, actual._rate)
        self.assertEqual(body, actual._body)

    def test_sync_form_mold(self):
        # Given
        form = _SyncRequestForm()
        envelope = _SyncRequest('sync_node', 'sync_lane', prio=99.1, rate=41.42, body=Text.create_from('Moo'))
        # When
        actual = form._mold(envelope)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(2, actual.size)
        self.assertEqual('sync', actual._tag)
        self.assertEqual('sync_node', actual.get_item(0).value.get_item(0).value.value)
        self.assertEqual('sync_lane', actual.get_item(0).value.get_item(1).value.value)
        self.assertEqual(99.1, actual.get_item(0).value.get_item(2).value.value)
        self.assertEqual(41.42, actual.get_item(0).value.get_item(3).value.value)
        self.assertEqual('Moo', actual.get_item(1).value)

    def test_sync_form_cast(self):
        # Given
        form = _SyncRequestForm()
        items = RecordMap.create()
        items.add(Slot.create_slot(Text.create_from('node'), Text.create_from('sync_node')))
        items.add(Slot.create_slot(Text.create_from('lane'), Text.create_from('sync_lane')))
        items.add(Slot.create_slot(Text.create_from('prio'), Num.create_from(33.12)))
        items.add(Slot.create_slot(Text.create_from('rate'), Num.create_from(12.33)))
        body = Attr.create_attr(Text.create_from('body'), Text.create_from('sync_body'))
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr(Text.create_from('sync'), items))
        record_map.add(body)
        # When
        actual = form._cast(record_map)
        # Then
        self.assertIsInstance(actual, _SyncRequest)
        self.assertEqual('sync_node', actual._node_uri)
        self.assertEqual('sync_lane', actual._lane_uri)
        self.assertEqual(33.12, actual._prio)
        self.assertEqual(12.33, actual._rate)
        self.assertEqual(body, actual._body.get_item(0))

    def test_link_form(self):
        # Given
        actual = _LinkRequestForm()
        # Then
        self.assertIsInstance(actual, _LinkRequestForm)
        self.assertEqual('link', actual._tag)

    def test_link_form_create_envelope(self):
        # Given
        form = _LinkRequestForm()
        body = Text.create_from('link_body')
        # When
        actual = form._create_envelope_from('moo', 'cow', prio=0.13, rate=0.26, body=body)
        # Then
        self.assertIsInstance(actual, _LinkRequest)
        self.assertEqual('moo', actual._node_uri)
        self.assertEqual('cow', actual._lane_uri)
        self.assertEqual(0.13, actual._prio)
        self.assertEqual(0.26, actual._rate)
        self.assertEqual(body, actual._body)

    def test_link_form_mold(self):
        # Given
        form = _LinkRequestForm()
        envelope = _LinkRequest('link_node', 'link_lane', prio=1, rate=2, body=Text.create_from('Moo'))
        # When
        actual = form._mold(envelope)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(2, actual.size)
        self.assertEqual('link', actual._tag)
        self.assertEqual('link_node', actual.get_item(0).value.get_item(0).value.value)
        self.assertEqual('link_lane', actual.get_item(0).value.get_item(1).value.value)
        self.assertEqual(1, actual.get_item(0).value.get_item(2).value.value)
        self.assertEqual(2, actual.get_item(0).value.get_item(3).value.value)
        self.assertEqual('Moo', actual.get_item(1).value)

    def test_link_form_cast(self):
        # Given
        form = _LinkRequestForm()
        items = RecordMap.create()
        items.add(Slot.create_slot(Text.create_from('node'), Text.create_from('link_node')))
        items.add(Slot.create_slot(Text.create_from('lane'), Text.create_from('link_lane')))
        items.add(Slot.create_slot(Text.create_from('prio'), Num.create_from(1)))
        items.add(Slot.create_slot(Text.create_from('rate'), Num.create_from(3)))
        body = Attr.create_attr(Text.create_from('body'), Text.create_from('link_body'))
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr(Text.create_from('link'), items))
        record_map.add(body)
        # When
        actual = form._cast(record_map)
        # Then
        self.assertIsInstance(actual, _LinkRequest)
        self.assertEqual('link_node', actual._node_uri)
        self.assertEqual('link_lane', actual._lane_uri)
        self.assertEqual(1, actual._prio)
        self.assertEqual(3, actual._rate)
        self.assertEqual(body, actual._body.get_item(0))

    def test_linked_form(self):
        # Given
        actual = _LinkedResponseForm()
        # Then
        self.assertIsInstance(actual, _LinkedResponseForm)
        self.assertEqual('linked', actual._tag)

    def test_linked_form_create_envelope(self):
        # Given
        form = _LinkedResponseForm()
        body = Text.create_from('linked_body')
        # When
        actual = form._create_envelope_from('boo', 'far', prio=1.13, rate=2.26, body=body)
        # Then
        self.assertIsInstance(actual, _LinkedResponse)
        self.assertEqual('boo', actual._node_uri)
        self.assertEqual('far', actual._lane_uri)
        self.assertEqual(1.13, actual._prio)
        self.assertEqual(2.26, actual._rate)
        self.assertEqual(body, actual._body)

    def test_linked_form_mold(self):
        # Given
        form = _LinkedResponseForm()
        envelope = _LinkedResponse('linked_node', 'linked_lane', prio=13, rate=15, body=Text.create_from('Foo'))
        # When
        actual = form._mold(envelope)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(2, actual.size)
        self.assertEqual('linked', actual._tag)
        self.assertEqual('linked_node', actual.get_item(0).value.get_item(0).value.value)
        self.assertEqual('linked_lane', actual.get_item(0).value.get_item(1).value.value)
        self.assertEqual(13, actual.get_item(0).value.get_item(2).value.value)
        self.assertEqual(15, actual.get_item(0).value.get_item(3).value.value)
        self.assertEqual('Foo', actual.get_item(1).value)

    def test_linked_form_cast(self):
        # Given
        form = _LinkedResponseForm()
        items = RecordMap.create()
        items.add(Slot.create_slot(Text.create_from('node'), Text.create_from('linked_node')))
        items.add(Slot.create_slot(Text.create_from('lane'), Text.create_from('linked_lane')))
        items.add(Slot.create_slot(Text.create_from('prio'), Num.create_from(14)))
        items.add(Slot.create_slot(Text.create_from('rate'), Num.create_from(13)))
        body = Attr.create_attr(Text.create_from('body'), Text.create_from('linked_body'))
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr(Text.create_from('linked'), items))
        record_map.add(body)
        # When
        actual = form._cast(record_map)
        # Then
        self.assertIsInstance(actual, _LinkedResponse)
        self.assertEqual('linked_node', actual._node_uri)
        self.assertEqual('linked_lane', actual._lane_uri)
        self.assertEqual(14, actual._prio)
        self.assertEqual(13, actual._rate)
        self.assertEqual(body, actual._body.get_item(0))

    def test_unlinked_form(self):
        # Given
        actual = _UnlinkedResponseForm()
        # Then
        self.assertIsInstance(actual, _UnlinkedResponseForm)
        self.assertEqual('unlinked', actual._tag)

    def test_unlinked_form_create_envelope(self):
        # Given
        form = _UnlinkedResponseForm()
        body = Text.create_from('unlinked_body')
        # When
        actual = form._create_envelope_from('oof', 'bar', prio=2, rate=6, body=body)
        # Then
        self.assertIsInstance(actual, _UnlinkedResponse)
        self.assertEqual('oof', actual._node_uri)
        self.assertEqual('bar', actual._lane_uri)
        self.assertEqual(2, actual._prio)
        self.assertEqual(6, actual._rate)
        self.assertEqual(body, actual._body)

    def test_unlinked_form_mold(self):
        # Given
        form = _UnlinkedResponseForm()
        envelope = _LinkedResponse('unlinked_node', 'unlinked_lane', prio=9, rate=10, body=Text.create_from('Baz'))
        # When
        actual = form._mold(envelope)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(2, actual.size)
        self.assertEqual('unlinked', actual._tag)
        self.assertEqual('unlinked_node', actual.get_item(0).value.get_item(0).value.value)
        self.assertEqual('unlinked_lane', actual.get_item(0).value.get_item(1).value.value)
        self.assertEqual(9, actual.get_item(0).value.get_item(2).value.value)
        self.assertEqual(10, actual.get_item(0).value.get_item(3).value.value)
        self.assertEqual('Baz', actual.get_item(1).value)

    def test_unlinked_form_cast(self):
        # Given
        form = _UnlinkedResponseForm()
        items = RecordMap.create()
        items.add(Slot.create_slot(Text.create_from('node'), Text.create_from('unlinked_node')))
        items.add(Slot.create_slot(Text.create_from('lane'), Text.create_from('unlinked_lane')))
        items.add(Slot.create_slot(Text.create_from('prio'), Num.create_from(20)))
        items.add(Slot.create_slot(Text.create_from('rate'), Num.create_from(25)))
        body = Attr.create_attr(Text.create_from('body'), Text.create_from('unlinked_body'))
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr(Text.create_from('unlinked'), items))
        record_map.add(body)
        # When
        actual = form._cast(record_map)
        # Then
        self.assertIsInstance(actual, _UnlinkedResponse)
        self.assertEqual('unlinked_node', actual._node_uri)
        self.assertEqual('unlinked_lane', actual._lane_uri)
        self.assertEqual(20, actual._prio)
        self.assertEqual(25, actual._rate)
        self.assertEqual(body, actual._body.get_item(0))

    def test_command_form(self):
        # Given
        actual = _CommandMessageForm()
        # Then
        self.assertIsInstance(actual, _CommandMessageForm)
        self.assertEqual('command', actual._tag)

    def test_command_form_create_envelope(self):
        # Given
        form = _CommandMessageForm()
        body = Text.create_from('message_body')
        # When
        actual = form._create_envelope_from('message_foo', 'message_bar', body=body)
        # Then
        self.assertIsInstance(actual, _CommandMessage)
        self.assertEqual('message_foo', actual._node_uri)
        self.assertEqual('message_bar', actual._lane_uri)
        self.assertEqual(body, actual._body)

    def test_command_form_mold(self):
        # Given
        form = _CommandMessageForm()
        envelope = _CommandMessage('command_node', 'command_lane', body=Text.create_from('Command_body'))
        # When
        actual = form._mold(envelope)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(2, actual.size)
        self.assertEqual('command', actual._tag)
        self.assertEqual('command_node', actual.get_item(0).value.get_item(0).value.value)
        self.assertEqual('command_lane', actual.get_item(0).value.get_item(1).value.value)
        self.assertEqual('Command_body', actual.get_item(1).value)

    def test_command_form_cast(self):
        # Given
        form = _CommandMessageForm()
        items = RecordMap.create()
        items.add(Slot.create_slot(Text.create_from('node'), Text.create_from('command_node')))
        items.add(Slot.create_slot(Text.create_from('lane'), Text.create_from('command_lane')))
        body = Attr.create_attr(Text.create_from('body'), Text.create_from('command_body'))
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr(Text.create_from('command'), items))
        record_map.add(body)
        # When
        actual = form._cast(record_map)
        # Then
        self.assertIsInstance(actual, _CommandMessage)
        self.assertEqual('command_node', actual._node_uri)
        self.assertEqual('command_lane', actual._lane_uri)
        self.assertEqual(body, actual._body.get_item(0))

    def test_event_form(self):
        # Given
        actual = _EventMessageForm()
        # Then
        self.assertIsInstance(actual, _EventMessageForm)
        self.assertEqual('event', actual._tag)

    def test_event_form_create_envelope(self):
        # Given
        form = _EventMessageForm()
        body = Text.create_from('event_body')
        # When
        actual = form._create_envelope_from('event_foo', 'event_bar', body=body)
        # Then
        self.assertIsInstance(actual, _EventMessage)
        self.assertEqual('event_foo', actual._node_uri)
        self.assertEqual('event_bar', actual._lane_uri)
        self.assertEqual(body, actual._body)

    def test_event_form_mold(self):
        # Given
        form = _EventMessageForm()
        envelope = _EventMessage('event_node', 'event_lane', body=Text.create_from('Event_body'))
        # When
        actual = form._mold(envelope)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(2, actual.size)
        self.assertEqual('event', actual._tag)
        self.assertEqual('event_node', actual.get_item(0).value.get_item(0).value.value)
        self.assertEqual('event_lane', actual.get_item(0).value.get_item(1).value.value)
        self.assertEqual('Event_body', actual.get_item(1).value)

    def test_event_form_cast(self):
        # Given
        form = _EventMessageForm()
        items = RecordMap.create()
        items.add(Slot.create_slot(Text.create_from('node'), Text.create_from('event_node')))
        items.add(Slot.create_slot(Text.create_from('lane'), Text.create_from('event_lane')))
        body = Attr.create_attr(Text.create_from('body'), Text.create_from('event_body'))
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr(Text.create_from('event'), items))
        record_map.add(body)
        # When
        actual = form._cast(record_map)
        # Then
        self.assertIsInstance(actual, _EventMessage)
        self.assertEqual('event_node', actual._node_uri)
        self.assertEqual('event_lane', actual._lane_uri)
        self.assertEqual(body, actual._body.get_item(0))



#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

from aiounittest import async_test

from swimai.recon._parsers import _OutputMessage
from swimai.recon._writers import _BoolWriter, _IdentWriter, _NumberWriter, _StringWriter, _SlotWriter, _ReconWriter, \
    _AttrWriter, _BlockWriter
from swimai.structures import Text, Slot, Attr, Num, Bool
from swimai.structures._structs import _Extant, _Absent, _Record
from test.utils import CustomItem


class TestWriters(unittest.TestCase):

    @async_test
    async def test_ident_writer_str(self):
        # Given
        message = 'hello'
        # When
        actual = await _IdentWriter._write(message)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('hello', actual._value)

    @async_test
    async def test_ident_writer_empty(self):
        # Given
        message = ''
        # When
        actual = await _IdentWriter._write(message)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('', actual._value)

    @async_test
    async def test_bool_writer_true(self):
        # Given
        message = True
        # When
        actual = await _BoolWriter._write(message)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('true', actual._value)

    @async_test
    async def test_bool_writer_false(self):
        # Given
        message = False
        # When
        actual = await _BoolWriter._write(message)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('false', actual._value)

    @async_test
    async def test_bool_writer_none(self):
        # Given
        message = None
        # When
        actual = await _BoolWriter._write(message)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('false', actual._value)

    @async_test
    async def test_number_writer_zero(self):
        # Given
        message = 0
        # When
        actual = await _NumberWriter._write(message)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('0', actual._value)

    @async_test
    async def test_number_writer_int(self):
        # Given
        message = 25
        # When
        actual = await _NumberWriter._write(message)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('25', actual._value)

    @async_test
    async def test_number_writer_float(self):
        # Given
        message = 0.02
        # When
        actual = await _NumberWriter._write(message)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('0.02', actual._value)

    @async_test
    async def test_number_writer_none(self):
        # Given
        message = None
        # When
        actual = await _NumberWriter._write(message)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('', actual._value)

    @async_test
    async def test_string_writer_str(self):
        # Given
        message = 'This is dog'
        # When
        actual = await _StringWriter._write(message)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('"This is dog"', actual._value)

    @async_test
    async def test_string_writer_empty(self):
        # Given
        message = None
        # When
        actual = await _StringWriter._write(message)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('""', actual._value)

    @async_test
    async def test_slot_writer_existing_key_and_value(self):
        # Given
        key = Text.create_from('animal')
        value = Text.create_from('dog')
        writer = _ReconWriter()
        # When
        actual = await _SlotWriter._write(key=key, writer=writer, value=value)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('animal:dog', actual._value)

    @async_test
    async def test_slot_writer_existing_key_missing_value(self):
        # Given
        key = Text.create_from('animal')
        writer = _ReconWriter()
        # When
        actual = await _SlotWriter._write(key=key, writer=writer)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('animal:', actual._value)

    @async_test
    async def test_slot_writer_missing_key_existing_value(self):
        # Given
        value = Text.create_from('dog')
        writer = _ReconWriter()
        # When
        actual = await _SlotWriter._write(value=value, writer=writer)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual(':dog', actual._value)

    @async_test
    async def test_slot_writer_missing_key_and_value(self):
        # Given
        writer = _ReconWriter()
        # When
        actual = await _SlotWriter._write(writer=writer)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual(':', actual._value)

    @async_test
    async def test_attr_writer_existing_key_and_value_text(self):
        # Given
        key = Text.create_from('bird')
        value = Text.create_from('chirp')
        writer = _ReconWriter()
        # When
        actual = await _AttrWriter._write(key=key, writer=writer, value=value)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('@bird(chirp)', actual._message)

    @async_test
    async def test_attr_writer_existing_key_and_value_slot(self):
        # Given
        key = Text.create_from('animal')
        value = _Record.create()
        value.add(Slot.create_slot(Text.create_from('dog'), Text.create_from('bark')))
        writer = _ReconWriter()
        # When
        actual = await _AttrWriter._write(key=key, writer=writer, value=value)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('@animal(dog:bark)', actual._message)

    @async_test
    async def test_attr_writer_missing_key_existing_value(self):
        # Given
        value = Text.create_from('chirp')
        writer = _ReconWriter()
        # When
        actual = await _AttrWriter._write(writer=writer, value=value)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('@(chirp)', actual._message)

    @async_test
    async def test_attr_writer_existing_key_missing_value(self):
        # Given
        key = Text.create_from('bird')
        writer = _ReconWriter()
        # When
        actual = await _AttrWriter._write(key=key, writer=writer)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('@bird', actual._message)

    @async_test
    async def test_attr_writer_existing_key_extant_value(self):
        # Given
        key = Text.create_from('bird')
        value = _Extant._get_extant()
        writer = _ReconWriter()
        # When
        actual = await _AttrWriter._write(key=key, writer=writer, value=value)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('@bird', actual._message)

    @async_test
    async def test_attr_writer_missing_key_and_value(self):
        # Given
        writer = _ReconWriter()
        # When
        actual = await _AttrWriter._write(writer=writer)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('@', actual._message)

    @async_test
    async def test_block_writer_attr(self):
        # Given
        items = list()
        items.append(Attr.create_attr(Text.create_from('dog'), Text.create_from('bark')))
        writer = _ReconWriter()
        # When
        actual = await _BlockWriter._write(items, writer=writer)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('@dog(bark)', actual._message)

    @async_test
    async def test_block_writer_text_single(self):
        # Given
        items = list()
        items.append(Text.create_from('Dead parrot'))
        writer = _ReconWriter()
        # When
        actual = await _BlockWriter._write(items, writer=writer)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('"Dead parrot"', actual._message)

    @async_test
    async def test_block_writer_text_multiple(self):
        # Given
        items = list()
        items.append(Text.create_from('foo_'))
        items.append(Text.create_from('bar'))
        writer = _ReconWriter()
        # When
        actual = await _BlockWriter._write(items, writer=writer)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('foo_bar', actual._message)

    @async_test
    async def test_block_writer_slot_single_first(self):
        # Given
        items = list()
        items.append(Slot.create_slot(Text.create_from('cat'), Text.create_from('meow')))
        writer = _ReconWriter()
        # When
        actual = await _BlockWriter._write(items, writer=writer, first=True)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('cat:meow', actual._message)

    @async_test
    async def test_block_writer_slot_single_not_first(self):
        # Given
        items = list()
        items.append(Slot.create_slot(Text.create_from('cat'), Text.create_from('meow')))
        writer = _ReconWriter()
        # When
        actual = await _BlockWriter._write(items, writer=writer, first=False)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual(',cat:meow', actual._message)

    @async_test
    async def test_block_writer_slot_multiple(self):
        # Given
        items = list()
        items.append(Slot.create_slot(Text.create_from('dog'), Text.create_from('bark')))
        items.append(Slot.create_slot(Text.create_from('cat'), Text.create_from('meow')))
        writer = _ReconWriter()
        # When
        actual = await _BlockWriter._write(items, writer=writer, first=True)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('dog:bark,cat:meow', actual._message)

    @async_test
    async def test_block_writer_slot_in_attr(self):
        # Given
        items = list()
        record_map = _Record.create()
        record_map.add(Slot.create_slot(Text.create_from('cat'), Text.create_from('meow')))
        items.append(Attr.create_attr(Text.create_from('animal'), record_map))
        writer = _ReconWriter()
        # When
        actual = await _BlockWriter._write(items, writer=writer, first=True)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('@animal(cat:meow)', actual._message)

    @async_test
    async def test_block_writer_slot_in_attr_and_slot(self):
        # Given
        items = list()
        record_map = _Record.create()
        record_map.add(Slot.create_slot(Text.create_from('dog'), Text.create_from('bark')))
        items.append(Attr.create_attr(Text.create_from('animal'), record_map))
        items.append(Slot.create_slot(Text.create_from('cat'), Text.create_from('meow')))
        writer = _ReconWriter()
        # When
        actual = await _BlockWriter._write(items, writer=writer, first=True)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('@animal(dog:bark){cat:meow}', actual._message)

    @async_test
    async def test_block_writer_multiple_attributes(self):
        # Given
        items = list()

        record_map = _Record.create()

        dog_map = _Record.create()
        dog_map.add(Slot.create_slot(Text.create_from('dog'), Text.create_from('bark')))
        record_map.add(Attr.create_attr(Text.create_from('Animal'), dog_map))

        cat_map = _Record.create()
        cat_map.add(Slot.create_slot(Text.create_from('cat'), Text.create_from('meow')))
        record_map.add(Attr.create_attr(Text.create_from('Animal'), cat_map))

        bird_map = _Record.create()
        bird_map.add(Slot.create_slot(Text.create_from('bird'), Text.create_from('chirp')))
        record_map.add(Attr.create_attr(Text.create_from('Animal'), bird_map))

        items.append(record_map)
        writer = _ReconWriter()
        # When
        actual = await _BlockWriter._write(items, writer=writer, first=True)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('@Animal(dog:bark)@Animal(cat:meow)@Animal(bird:chirp)', actual._message)

    @async_test
    async def test_block_writer_nested_attributes(self):
        # Given
        items = list()
        name_record = _Record.create()
        name_record.add(Slot.create_slot(Text.create_from('Name'), Text.create_from('Collie')))
        breed_record = _Record.create()
        breed_record.add(Attr.create_attr(Text.create_from('Breed'), name_record))
        dog_record = _Record.create()
        dog_record.add(Slot.create_slot(Text.create_from('Dog'), breed_record))
        species_record = _Record.create()
        species_record.add(Attr.create_attr(Text.create_from('Species'), dog_record))
        animals_record = _Record.create()
        animals_record.add(Slot.create_slot(Text.create_from('Animals'), species_record))
        items.append(Attr.create_attr(Text.create_from('Zoo'), animals_record))
        writer = _ReconWriter()
        # When
        actual = await _BlockWriter._write(items, writer=writer, first=True)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('@Zoo(Animals:@Species(Dog:@Breed(Name:Collie)))', actual._message)

    @async_test
    async def test_block_writer_empty(self):
        # Given
        items = list()
        writer = _ReconWriter()
        # When
        actual = await _BlockWriter._write(items, writer=writer)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('', actual._message)

    @async_test
    async def test_write_record_single(self):
        # Given
        record = _Record.create()
        record.add(Text.create_from('Dog'))
        writer = _ReconWriter()
        # When
        actual = await writer._write_record(record)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual(3, actual._size)
        self.assertEqual('Dog', actual._value)
        self.assertEqual('g', actual._last_char)

    @async_test
    async def test_write_record_multiple(self):
        # Given
        record = _Record.create()
        record.add(Text.create_from('Dog'))
        record.add(Text.create_from('Cat'))
        writer = _ReconWriter()
        # When
        actual = await writer._write_record(record)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual(6, actual._size)
        self.assertEqual('DogCat', actual._value)
        self.assertEqual('t', actual._last_char)

    @async_test
    async def test_write_record_empty(self):
        # Given
        record = _Record.create()
        writer = _ReconWriter()
        # When
        actual = await writer._write_record(record)
        # Then
        self.assertIsNone(actual)

    @async_test
    async def test_write_value_record(self):
        # Given
        record = _Record.create()
        record.add(Slot.create_slot(Text.create_from('Cow'), Text.create_from('Moo')))
        writer = _ReconWriter()
        # When
        actual = await writer._write_value(record)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual(7, actual._size)
        self.assertEqual('Cow:Moo', actual._value)
        self.assertEqual('o', actual._last_char)

    @async_test
    async def test_write_value_text_ident(self):
        # Given
        ident = Text.create_from('Duck')
        writer = _ReconWriter()
        # When
        actual = await writer._write_value(ident)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual(4, actual._size)
        self.assertEqual('Duck', actual._value)
        self.assertEqual('k', actual._last_char)

    @async_test
    async def test_write_value_text_string(self):
        # Given
        string = Text.create_from('$duck')
        writer = _ReconWriter()
        # When
        actual = await writer._write_value(string)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual(7, actual._size)
        self.assertEqual('"$duck"', actual._value)
        self.assertEqual('"', actual._last_char)

    @async_test
    async def test_write_value_number(self):
        # Given
        number = Num.create_from(-13.1)
        writer = _ReconWriter()
        # When
        actual = await writer._write_value(number)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual(5, actual._size)
        self.assertEqual('-13.1', actual._value)
        self.assertEqual('1', actual._last_char)

    @async_test
    async def test_write_value_bool(self):
        # Given
        boolean = Bool.create_from(False)
        writer = _ReconWriter()
        # When
        actual = await writer._write_value(boolean)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual(5, actual._size)
        self.assertEqual('false', actual._value)
        self.assertEqual('e', actual._last_char)

    @async_test
    async def test_write_value_absent(self):
        # Given
        absent = _Absent._get_absent()
        writer = _ReconWriter()
        # When
        actual = await writer._write_value(absent)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual(0, actual._size)
        self.assertEqual('', actual._value)

    @async_test
    async def test_write_slot(self):
        # Given
        key = Text.create_from('Hello')
        value = Text.create_from('Friend')
        writer = _ReconWriter()
        # When
        actual = await writer._write_slot(key, value)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual(12, actual._size)
        self.assertEqual('Hello:Friend', actual._value)
        self.assertEqual('d', actual._last_char)

    @async_test
    async def test_write_attr(self):
        # Given
        key = Text.create_from('Hello')
        value = Text.create_from('Friend')
        writer = _ReconWriter()
        # When
        actual = await writer._write_attr(key, value)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual(14, actual._size)
        self.assertEqual('@Hello(Friend)', actual._value)
        self.assertEqual(')', actual._last_char)

    @async_test
    async def test_write_item_attr(self):
        # Given
        item = Attr.create_attr(Text.create_from('Cat'), Text.create_from('Meow'))
        writer = _ReconWriter()
        # When
        actual = await writer._write_item(item)
        # Then
        self.assertIsInstance(actual, str)
        self.assertEqual('@Cat(Meow)', actual)

    @async_test
    async def test_write_item_slot(self):
        # Given
        item = Slot.create_slot(Text.create_from('Age'), Num.create_from(32))
        writer = _ReconWriter()
        # When
        actual = await writer._write_item(item)
        # Then
        self.assertIsInstance(actual, str)
        self.assertEqual('Age:32', actual)

    @async_test
    async def test_write_item_value(self):
        # Given
        item = Text.create_from('Horse#')
        writer = _ReconWriter()
        # When
        actual = await writer._write_item(item)
        # Then
        self.assertIsInstance(actual, str)
        self.assertEqual('"Horse#"', actual)

    @async_test
    async def test_write_item_invalid(self):
        # Given
        item = CustomItem()
        writer = _ReconWriter()
        # When
        with self.assertRaises(TypeError) as error:
            await writer._write_item(item)

        # Then
        message = error.exception.args[0]
        self.assertEqual('No Recon serialization for CustomItem!', message)
